#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

@author: Gasmi
"""

import matplotlib.pyplot as plt
import volmdlr.core_compiled 
import volmdlr
import volmdlr.core
from volmdlr.core_compiled import Matrix33

from itertools import combinations
import numpy as npy
import volmdlr.edges 
import volmdlr.wires 
import volmdlr.faces 
from volmdlr.core_compiled import Matrix33
import math
from dessia_common import DessiaObject
from typing import TypeVar, List, Tuple,Dict
import matplotlib
import random
from itertools import product 
from matplotlib.colors import LinearSegmentedColormap
cdict = {'red':  [(0.0, 0.0, 0.0),
                   (1.0, 1.0, 1.0)],
         'green': [(0.0, 0.0, 0.0),
                   (1.0, 0.0, 0.0)],
         'blue':  [(0.0, 1.0, 1.0),
                   (1.0, 0.0, 0.0)]}
blue_red = LinearSegmentedColormap('BLueRed', cdict)

class FlatElementError(Exception):
    pass

def find_duplicate_linear_element(linear_elements1, linear_elements2):
    duplicates = []
    for linear_element in linear_elements1:
        if linear_element in linear_elements2 and linear_element not in duplicates:
            duplicates.append(linear_element)
    return duplicates


class LinearElement(volmdlr.edges.LineSegment2D):
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True
    def __init__(self, start:volmdlr.Point2D,end:volmdlr.Point2D, interior_normal:volmdlr.Vector2D,name=''):
        self.points=[start, end]
        self.interior_normal = interior_normal
        
        volmdlr.edges.LineSegment2D.__init__(self,start=start,end=end,name=name)
        
    def __hash__(self):
        return self.start.__hash__() + self.end.__hash__()
        
    def __eq__(self, other_linear_element):
        if self.__class__ != other_linear_element.__class__:
            return False
        return (self.start == other_linear_element.start and self.end == other_linear_element.end) \
            or (self.start== other_linear_element.end and self.end == other_linear_element.start)
            
    
    def plot(self, ax=None, color='k', width=None, plot_points=False):
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        if width is None:
            width=1
        if plot_points:
            ax.plot([self.start.x, self.end.x], [self.start.y, self.end.y], color=color, marker='o', linewidth=width)
        else:
            ax.plot([self.start.x, self.end.x], [self.start.y, self.end.y], color=color, linewidth=width)
        return ax


class TriangularElement(volmdlr.wires.Triangle2D):
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points):
        self.points = points
        self.linear_elements = self._to_linear_elements()
        self.form_functions = self._form_functions()
        
        self.center = (self.points[0]+self.points[1]+self.points[2])/3
        
        self.area = self._area()
        
        volmdlr.wires.Triangle2D.__init__(self,points=points, name='')
        
    def _to_linear_elements(self):
        vec1 = volmdlr.Vector2D(self.points[1].x - self.points[0].x,
                                self.points[1].y - self.points[0].y)
        vec2 = volmdlr.Vector2D(self.points[2].x - self.points[1].x,
                                self.points[2].y - self.points[1].y)
        vec3 = volmdlr.Vector2D(self.points[0].x - self.points[2].x,
                                self.points[0].y - self.points[2].y)
        normal1 = volmdlr.Vector2D(-vec1.y, vec1.x)
        normal2 = volmdlr.Vector2D(-vec2.y, vec2.x)
        normal3 = volmdlr.Vector2D(-vec3.y, vec3.x)
        normal1.normalize()
        normal2.normalize()
        normal3.normalize()
        if normal1.dot(vec2) < 0:
            normal1 = - normal1
        if normal2.dot(vec3) < 0:
            normal2 = - normal2
        if normal3.dot(vec1) < 0:
            normal3 = - normal3
        linear_element_1 = LinearElement(self.points[0], self.points[1],
                                         normal1)
        linear_element_2 = LinearElement(self.points[1], self.points[2],
                                         normal2)
        linear_element_3 = LinearElement(self.points[2], self.points[0],
                                         normal3)
        return [linear_element_1, linear_element_2, linear_element_3]
    
    def _form_functions(self):
        a = Matrix33(1, self.points[0].x, self.points[0].y,
                     1, self.points[1].x, self.points[1].y,
                     1, self.points[2].x, self.points[2].y)
        try:
            inv_a = a.inverse()
        except ValueError:
            self.plot()
            print('buggy element area', self.area)
            raise FlatElementError('form function bug')
        x1 = inv_a.vector_multiplication(volmdlr.X3D)
        x2 = inv_a.vector_multiplication(volmdlr.Y3D)
        x3 = inv_a.vector_multiplication(volmdlr.Z3D)
       
        return x1, x2, x3

    # def _quadratic_form_functions(self):
    #     a = [[1, self.points[0][0], self.points[0][1],self.points[0][0]**2,self.points[0][0]*self.points[0][1],self.points[0][1]**2],
    #           [1, self.points[1][0], self.points[1][1],self.points[1][0]**2,self.points[1][0]*self.points[1][1],self.points[1][1]**2],
    #           [1, self.points[2][0], self.points[2][1],self.points[2][0]**2,self.points[2][0]*self.points[2][1],self.points[2][1]**2],
    #           [1, self.points[3][0], self.points[3][1],self.points[3][0]**2,self.points[3][0]*self.points[3][1],self.points[3][1]**2],
    #           [1, self.points[4][0], self.points[4][1],self.points[4][0]**2,self.points[4][0]*self.points[4][1],self.points[4][1]**2],
    #           [1, self.points[5][0], self.points[5][1],self.points[5][0]**2,self.points[5][0]*self.points[5][1],self.points[5][1]**2]]

    #     try :
    #         inv_a = a.inverse()
    #     except ValueError:
    #         self.plot()
    #         print(self._area())
    #         raise FlatElementError('form function bug')
    #     x1 = inv_a.dot([1,0,0,0,0,0])
    #     x2 = inv_a.dot([1,0,0,0,0,0])
    #     x3 = inv_a.dot([1,0,0,0,0,0])
    #     x4=inv_a.dot([1,0,0,0,0,0])
        
    #     return x1, x2, x3

    def _area(self):
        u = self.points[1] - self.points[0]
        v = self.points[2] - self.points[0]
        return abs(u.cross(v)) / 2

    def point_belongs(self, point):
        polygon = volmdlr.wires.ClosedPolygon2D(self.points)
        point_belongs = polygon.point_belongs(point)
        return point_belongs

    def rotation(self, center, angle, copy=True):
        if copy:
            return TriangularElement([pt.rotation(center, angle, copy=True)
                                      for pt in self.points])
        else:
            for pt in self.points:
                pt.Rotation(center, angle, copy=False)
                
    def translation(self, offset, copy=True):
        if copy:
            return TriangularElement([pt.translation(offset, copy=True)
                                      for pt in self.points])
        else:
            for pt in self.points:
                pt.translation(offset, copy=False)
                
    def axial_symmetry(self, line, copy=True):
        p1, p2 = line.points
        symmetric_points = []
        for point in self.points:
            u = p2 - p1
            t = (point-p1).dot(u) / u.norm()**2
            projection = p1 + t * u
            symmetric_point = volmdlr.Point2D(*(2 * projection - point))
            symmetric_points.append(symmetric_point)
        if copy: 
            return TriangularElement(symmetric_points)
        else:
            for i, point in enumerate(self.points):
                point = symmetric_points[i]
   
    def plot(self, ax=None, color='k', width=None,
             plot_points=False, fill=False):
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
            
        if fill:
            x = [p[0] for p in self.points]
            y = [p[1] for p in self.points]
            plt.fill(x, y, facecolor=color, edgecolor="k")
            return ax
        
        for p1, p2 in zip(self.points, self.points[1:]+[self.points[0]]):
            if width is None:
                width = 1
            if plot_points:
                ax.plot([p1.x, p2.x], [p1.y, p2.y], color=color,
                        marker='o', linewidth=width)
            else:
                ax.plot([p1.x, p2.x], [p1.y, p2.y], color=color,
                        linewidth=width)
        return ax

    def triangle_to_polygon(self):
        points = self.points
        return volmdlr.wires.ClosedPolygon2D(points)
       

class ElementsGroup(DessiaObject):
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, elements:List[TriangularElement], name:str):
        self.elements = elements
        self.name = name

        DessiaObject.__init__(self, name=name)

    def point_to_element(self, point):
        for element in self.elements:
            if element.point_belongs(point):
                return element
        return None

    def rotation(self, center, angle, copy=True):
        if copy:
            return Mesh([elem.rotation(center, angle, copy=True)
                         for elem in self.elements])
        else:
            for elem in self.elements:
                elem.rotation(center, angle, copy=False)
                
    def translation(self, offset, copy=True):
        if copy:
            return Mesh([elem.translation(offset, copy=True)
                         for elem in self.elements])
        else:
            for elem in self.elements:
                elem.translation(offset, copy=False)
                
    def plot(self, ax=None, color='k', fill=False):
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        for element in self.elements:
            element.plot(ax=ax, color=color, fill=fill)
        return ax
        

class Mesh(DessiaObject):
    _standalone_in_db = True
    _non_serializable_attributes = ['node_to_index']
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, elements_groups: List[ElementsGroup]):
        self.elements_groups = elements_groups
        self.nodes = self._set_nodes_number()
        self.node_to_index = {self.nodes[i]: i for i in range(len(self.nodes))}
        
        DessiaObject.__init__(self, name='')

    def __add__(self, other_mesh):
        new_nodes = self.nodes[:]
        new_nodes_index = {p: i for i, p in enumerate(self.points)}
        ip = len(new_nodes)
        for point in other_mesh.nodes:
            if point not in new_nodes_index:
                new_nodes_index[point] = ip
                ip += 1
                new_nodes.append(point)

        new_elements_groups = self.elements_groups[:]
        for i1, i2, i3 in other_mesh.elements_groups:
            p1 = other_mesh.nodes[i1]
            p2 = other_mesh.nodes[i2]
            p3 = other_mesh.nodes[i3]
            new_elements_groups.append((new_nodes_index[p1],
                                        new_nodes_index[p2],
                                        new_nodes_index[p3]))
        return self.__class__(new_elements_groups)

    def _set_nodes_number(self):
        nodes = set()
        for elements_group in self.elements_groups:
            for element in elements_group.elements:
                nodes.add(element.points[0])
                nodes.add(element.points[1])
                nodes.add(element.points[2])
        return tuple(nodes)

    def point_to_element(self, point):
        for element_group in self.elements_groups:
            element = element_group.point_to_element(point)
            if element is not None:
                return element
        return None
    
    def set_node_displacement_index(self):
        indexes = {}
        for node in self.nodes:
            indexes[node] = [2*self.node_to_index[node],
                             2*self.node_to_index[node]+1]
        return indexes
    
    def boundary_dict(self):
        boundary_dict = {}
        for elements_group1, elements_group2 in combinations(
                self.elements_groups, 2):
            linear_elements1 = []
            linear_elements2 = []
            for element in elements_group1.elements:
                linear_elements1.extend(element.linear_elements)
            for element in elements_group2.elements:
                linear_elements2.extend(element.linear_elements)
            duplicate_linear_elements = find_duplicate_linear_element(
                linear_elements1, linear_elements2)
            if duplicate_linear_elements:
                boundary_dict[(elements_group1,
                               elements_group2)] = duplicate_linear_elements
        return boundary_dict
   
    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        for elements_group in self.elements_groups:
            elements_group.plot(ax=ax)
        return ax
    
    def plot_data(self, pos=0, quote=True, constructor=True, direction=1):
        plot_datas = []
        for element_group in self.elements_groups:
            for element in element_group.elements:
                c1 = volmdlr.wires.Contour2D([volmdlr.edges.LineSegment2D(
                    element.points[0], element.points[1])])
                c2 = volmdlr.wires.Contour2D([volmdlr.edges.LineSegment2D(
                    element.points[1], element.points[2])])
                c3 = volmdlr.wires.Contour2D([volmdlr.edges.LineSegment2D(
                    element.points[2], element.points[0])])
                plot_datas.append(c1.plot_data())
                plot_datas.append(c2.plot_data())
                plot_datas.append(c3.plot_data())
                # plot_datas.extend([c1, c2, c3])
        return plot_datas

    def plot_displaced_mesh(self,
                            node_displacement: Dict[volmdlr.Point2D,
                                                    List[float]],
                            ax=None, amplification=0.5):

        deformed_mesh = self.copy()
        nodes = deformed_mesh.nodes
  
        for node in nodes:
            for displaced_node in node_displacement:
                if node == displaced_node:
                    node.x += amplification*node_displacement[
                        displaced_node][0]
                    node.y += amplification*node_displacement[
                        displaced_node][1]
            
        ax = deformed_mesh.plot(ax=ax) 
        ax.set_aspect('equal')           
    
        return ax


class Mesher(DessiaObject):
    
    def __init__(self,interior_contours:List[volmdlr.wires.Contour2D],exterior_contours:List[volmdlr.wires.Contour2D],triangles:List[TriangularElement],nodes_len:float):
        self.nodes_len=nodes_len
        self.interior_contours=interior_contours
        self.exterior_contours=exterior_contours
        self.triangles=triangles
    
        
    def  neighbour_edge(self,n:int,i:int,di:int):
        return (i+di)%n  
   
    
    def edge_max_distance(self,polygone:volmdlr.wires.ClosedPolygon2D,P0:volmdlr.Point2D,P1:volmdlr.Point2D,P2:volmdlr.Point2D,indexes:List[float]):
        n=len(polygone.points)
        distance=0
        j=None
        triangle=volmdlr.wires.Triangle2D([P0,P1,P2])
        for i in range(n):
            if not (i in indexes):
                M=polygone.points[i]
                if triangle.is_inside_triangle(M):
                    d=abs(triangle.line_equation(P1,P2,M))
                    if d > distance:
                        distance=d
                        j=i
       
        return j
    
    
    
    
    def left_edge(self,polygone:volmdlr.wires.ClosedPolygon2D):
        n=len(polygone.points)
        x=polygone.points[0].x
        j=0
        for i in range (1,n):
            if polygone.points[i].x < x:
                x=polygone.points[i].x
                j=i
        return j
            
    def new_polygon(self,polygone:volmdlr.wires.ClosedPolygon2D,i_beg:int,i_end=int):
        n=len(polygone.points)
        u=[]
        i=i_beg
        while i!=i_end:
            u.append(polygone.points[i])
            i=self.neighbour_edge(n,i,1)
        u.append(polygone.points[i_end])
        p=volmdlr.wires.ClosedPolygon2D(u)
        
        return p
    
    def triangulation_polygone_recursive(self,polygone:volmdlr.wires.ClosedPolygon2D,triangles:List[volmdlr.wires.Triangle2D]):
        
        n=len(polygone.points)
        j0=self.left_edge(polygone)
      
       
        j1=self.neighbour_edge(n,j0,1)
       
        j2=self.neighbour_edge(n,j0,-1)
        
        P0=polygone.points[j0]
        P1=polygone.points[j1]
        P2=polygone.points[j2]
        j=self.edge_max_distance(polygone,P0,P1,P2,[j0,j1,j2])
        
        if j==None:
            triangle=volmdlr.wires.Triangle2D([P0,P1,P2])
            triangles.append(triangle)
            polygone_1=self.new_polygon(polygone,j1,j2)
                            
            if len(polygone_1.points)==3:
                    
                    new_triangle=volmdlr.wires.Triangle2D([polygone_1.points[0],polygone_1.points[1],
                                                polygone_1.points[2]])    
                    triangles.append(new_triangle)  
                                          
            else :
             
                  self.triangulation_polygone_recursive(polygone_1,triangles)
     
        else : 
            
            polygone_1=self.new_polygon(polygone,j0,j)
            polygone_2=self.new_polygon(polygone,j,j0)    
            
            if len(polygone_1.points)==3:
                new_triangle=volmdlr.wires.Triangle2D([polygone_1.points[0],polygone_1.points[1],
                                            polygone_1.points[2]])                
                triangles.append(new_triangle)
               
            else :
                self.triangulation_polygone_recursive(polygone_1,triangles)
                
            if len(polygone_2.points)==3:
                new_triangle=volmdlr.wires.Triangle2D([polygone_2.points[0],polygone_2.points[1],
                                            polygone_2.points[2]])                    
                triangles.append(new_triangle)

            else :
                    
                self.triangulation_polygone_recursive(polygone_2,triangles)
            
        return triangles 
    
    
    
    def _is_convex(self,p1:volmdlr.Point2D, p2:volmdlr.Point2D, p3:volmdlr.Point2D):
        return self._triangle_sum(p1.x, p1.y, p2.x, p2.y, p3.x,p3.y) < 0
    
    def _is_clockwise(self,polygon:volmdlr.wires.ClosedPolygon2D):
        s = 0
        polygon_count = len(polygon.points)
        for i in range(polygon_count):
            point = polygon.points[i]
            point2 = polygon.points[(i + 1) % polygon_count]
            s += (point2.x - point.x) * (point2.y + point.y)
        return s > 0
    
    def _triangle_sum(self,x1, y1, x2, y2, x3, y3):
        return x1 * (y3 - y2) + x2 * (y1 - y3) + x3 * (y2 - y1)
    
    def _contains_no_points(self,p1:volmdlr.Point2D,p2:volmdlr.Point2D,p3:volmdlr.Point2D, polygon:volmdlr.wires.ClosedPolygon2D):
       triangle=volmdlr.wires.Triangle2D([p1,p2,p3])
       for pn in polygon.points:
            if pn in [p1, p2, p3]:
                continue
            elif triangle.is_inside_triangle(pn):
                return False
       return True

    def _is_ear(self,p1:volmdlr.Point2D,p2:volmdlr.Point2D,p3:volmdlr.Point2D, polygon:volmdlr.wires.ClosedPolygon2D):
        triangle=volmdlr.wires.Triangle2D([p1,p2,p3])
        ear = self._contains_no_points(p1, p2, p3, polygon) and \
            self._is_convex(p1, p2, p3) and \
            triangle.area > 0
        return ear
    
    def earclip(self,polygon:volmdlr.wires.ClosedPolygon2D):
       
        possible_triangles=[]
        ear_vertex =[]
       
       
        if self._is_clockwise(polygon):
            polygon.points.reverse()
    
        point_count = len(polygon.points)
       
         
        for i in range(point_count)  :
                   
            prev_index = i - 1
            prev_point = polygon.points[prev_index]
          
            point = polygon.points[i]
            next_index = (i + 1) % point_count
            next_point = polygon.points[next_index]
    
            if self._is_ear(prev_point, point, next_point, polygon):
                ear_vertex.append(point)
                    
        while ear_vertex and point_count >=3  :
            ear = ear_vertex.pop(0)
            i = polygon.points.index(ear)
            prev_index = i - 1
            prev_point = polygon.points[prev_index]
            next_index = (i + 1) % point_count
            next_point = polygon.points[next_index]
            prev_prev_point = polygon.points[prev_index - 1]
            next_next_index = (i + 1) % point_count
            next_next_point = polygon.points[next_next_index]
            polygon.points.remove(ear)
            point_count -= 1
            
            p1=volmdlr.Point2D(prev_point[0], prev_point[1])
            p2=volmdlr.Point2D(ear[0], ear[1])
            p3= volmdlr.Point2D(next_point[0], next_point[1])
            triangle=volmdlr.wires.Triangle2D([p1,p2,p3])  
            possible_triangles.append(triangle)
             
            if point_count > 3:
                prev_prev_point = polygon.points[prev_index - 1]
                next_next_index = (i + 1) % point_count
                next_next_point = polygon.points[next_next_index]
    
                groups = [
                    (prev_prev_point, prev_point, next_point, polygon),
                    (prev_point, next_point, next_next_point, polygon),
                ]
                for group in groups:
                    p = group[1]
                    if self._is_ear(*group):
                        if p not in ear_vertex:
                            ear_vertex.append(p)
                    elif p in ear_vertex:
               
                         ear_vertex.remove(p)
                           
        return possible_triangles   
                 
                                 
        
    def basic_triangulation(self,polygon1:volmdlr.wires.ClosedPolygon2D,polygon2:volmdlr.wires.ClosedPolygon2D,
                           segment_to_nodes:Dict[volmdlr.edges.LineSegment2D,List[volmdlr.Point2D]]):
        triangles=[]
        
        for j in range(len(polygon1.line_segments)):
                    
            pj=segment_to_nodes[polygon1.line_segments[j]]
            qj=segment_to_nodes[polygon2.line_segments[j]]
            u=len(pj)
            v=len(qj)
            if u==2 and v==2 : 
                new_triangle_1=volmdlr.wires.Triangle2D([pj[0],pj[1],qj[0]])
                triangles.append(new_triangle_1)                          
                new_triangle_2=volmdlr.wires.Triangle2D([pj[1],qj[1],qj[0]])
                triangles.append(new_triangle_2)                           
            if u>=v:
                for i in range(v-1):
                    new_triangle_1=volmdlr.wires.Triangle2D([pj[i+1],pj[i],qj[i]])
                   
                    triangles.append(new_triangle_1)
                    new_triangle_2=volmdlr.wires.Triangle2D([qj[i+1],pj[i+1],qj[i]])
                   
                    triangles.append(new_triangle_2)
                              
                for  i in range(v-1,u-1):
                                
                     new_triangle=volmdlr.wires.Triangle2D([pj[i],qj[v-1],pj[i+1]])
                                     
                     triangles.append(new_triangle)  
            else :
                  for i in range(u-1):
                    new_triangle_1=volmdlr.wires.Triangle2D([qj[i+1],qj[i],pj[i]])
                   
                    triangles.append(new_triangle_1)
                    new_triangle_2=volmdlr.wires.Triangle2D([pj[i+1],qj[i+1],pj[i]])
                   
                    triangles.append(new_triangle_2)
                              
                  for  i in range(u-1,v-1):
                                
                     new_triangle=volmdlr.wires.Triangle2D([qj[i],pj[u-1],qj[i+1]])
                                     
                     triangles.append(new_triangle)  
                             
        return triangles                            
    
    def alternative_triangulation(self,polygon:volmdlr.wires.ClosedPolygon2D,interior_polygon:volmdlr.wires.ClosedPolygon2D,empty:bool,far:bool):
        # ax=plt.subplot()
        all_aspect_ratios=[]
        segment_to_nodes={}
        
        all_offsets=[]
        all_meshes=[]
        offset_values=[]
        good_meshes=[]
        p=9
        k=3
        xmin,xmax,ymin,ymax=polygon.bounding_rectangle()

     
        offset_len=min(xmax-xmin, ymax-ymin)

        
        while k<12:
            repair = False
            polygon_offsets=[]
            polygon_offsets+=[polygon]
            while repair is False :
                
                new_polygon=polygon.offset(-p*offset_len/(10*k))
                
                if not new_polygon.self_intersects()[0] :
                    if new_polygon.area() > polygon.area():
                        good_offset=polygon.offset(p*offset_len/(10*k))
                        if not good_offset.self_intersects()[0]:
                           polygon_offsets.append(good_offset)
                             
                        else :
                            
                        
                            offset_values.append(p*offset_len/(10*k))
                        
                    else :

                         offset_values.append(-p*offset_len/(10*k))
                         polygon_offsets.append(new_polygon)
                    xmin_2,xmax_2,ymin_2,ymax_2=polygon_offsets[-1].bounding_rectangle()
                    offset_len2=min(xmax_2-xmin_2, ymax_2-ymin_2)
                    offset2=polygon_offsets[-1].offset(-p*offset_len2/(10*k))
            
                    if offset2.self_intersects()[0]:
                        polygon_offsets.append(offset2.select_reapaired_polygon([]))
                    else:
                        if offset2.area() > polygon_offsets[-1].area() :
                            good_offset_2=polygon_offsets[-1].offset(p*offset_len2/(10*k))
                          
                            if not good_offset_2.self_intersects()[0]: 
                                polygon_offsets.append(good_offset_2) 
                            else :
                                polygon_offsets.append(good_offset_2.select_reapaired_polygon([])) 
                        else :                           
                            polygon_offsets.append(offset2)      

                            offset_values.append(-p*offset_len/(10*k))
                            polygon_offsets.append(new_polygon)
                        
                    xmin_2,xmax_2,ymin_2,ymax_2=polygon_offsets[-1].bounding_rectangle()
                    offset_len2=min(xmax_2-xmin_2, ymax_2-ymin_2)
                    offset2=polygon_offsets[-1].offset(-p*offset_len2/(10*k))
            
                    if offset2.self_intersects()[0]:
                        polygon_offsets.append(offset2.select_reapaired_polygon([]))
                    else:
                        if offset2.area() > polygon_offsets[-1].area() :
                            good_offset_2=polygon_offsets[-1].offset(p*offset_len2/(10*k))
                          
                            if not good_offset_2.self_intersects()[0]: 
                                polygon_offsets.append(good_offset_2) 
                            else :
                                polygon_offsets.append(good_offset_2.select_reapaired_polygon([])) 
                        else :                           
                            polygon_offsets.append(offset2)    

                    repair=True
                    
                    
                if new_polygon.self_intersects()[0] :
                    
                    # polygon_offsets.append(new_polygon.select_reapaired_polygon([]))
                    # new_polygon.select_reapaired_polygon([]).plot()
                    # new_polygon.select_reapaired_polygon([]).MPLPlot()
                
                    
                    # rec2=polygon_offsets[-1].bounding_rectangle2()
                    # offset_len2=rec2.min_length()
                    # offset2=polygon_offsets[-1].offset(-p*offset_len2/(10*k))
                    
                    # if offset2.self_intersects()[0]:
                    #     polygon_offsets.append(offset2.select_reapaired_polygon([]))
                       
                    # else:
                    #     if offset2.area> polygon_offsets[-1].area :
                    #         polygon_offsets.append(p*offset_len2/(10*k)) 
                    #     else :                           
                    #         polygon_offsets.append(offset2)
                          
                    repair=True
             
            all_offsets.append(polygon_offsets)        
            k=k+1
                    
        for polygon_offsets in all_offsets:

            offset_triangles=[]  
            for polygon in polygon_offsets:
               
                for segment in polygon.line_segments:
                    segment_to_nodes[segment]=segment.discretise(self.nodes_len)
                    
                    
            if len(polygon_offsets)>2:      
                for k in range(len(polygon_offsets)-2):
                
                    if len(polygon_offsets[k].line_segments)==len(polygon_offsets[k+1].line_segments):
                        offset_triangles+=self.basic_triangulation(polygon_offsets[k],
                                                            polygon_offsets[k+1],segment_to_nodes)
                  
                    else :
                    
                        offset_triangles+=self.mesh_in_between(polygon_offsets[k+1],
                                                            polygon_offsets[k],True)
                    
                     
            l=len(polygon_offsets)
          
            if len(polygon_offsets[-1].line_segments)==len(polygon_offsets[l-2].line_segments):
               offset_triangles+=self.basic_triangulation(polygon_offsets[l-2],polygon_offsets[-1],segment_to_nodes)
               if  not empty :
                    
                    last_points=[]
                    for segment in polygon_offsets[-1].line_segments:
                        for point in segment_to_nodes[segment]:
                            if point not in last_points:
                                last_points.append(point)
                    last_polygon=volmdlr.wires.ClosedPolygon2D(last_points)
                    ear=self.earclip(last_polygon)
                    for triangle in ear : 
                     
                        if triangle.area>10e-9:
                            offset_triangles.append(triangle)
               else :
                    if far :
                        offset_triangles+=self.mesh_in_between(interior_polygon,polygon_offsets[-1],True)
                               
     
            else:
                offset_triangles+=self.mesh_in_between(polygon_offsets[-1],polygon_offsets[l-2],empty)                
                          
            all_meshes.append(offset_triangles)  
        
        
        for mesh in all_meshes:
            if self.triangulation_max_aspect_ratio(mesh)!=0:
                
                all_aspect_ratios.append(self.triangulation_max_aspect_ratio(mesh))
                good_meshes.append(mesh)
    
        index=all_aspect_ratios.index(min(all_aspect_ratios))
        return good_meshes[index]

    def mesh_in_between(self,in_polygon:volmdlr.wires.ClosedPolygon2D,
                        out_polygon:volmdlr.wires.ClosedPolygon2D,empty:bool):
        
        # ax=plt.subplot()
        projection_points=[]
        segment_to_nodes={}
        closest_segment={}
        all_triangles=[]
        out_point_image={}

        for segment in out_polygon.line_segments:
            segment_to_nodes[segment]=segment.discretise(self.nodes_len)
         
                
        for segment in out_polygon.line_segments:
            for point in segment_to_nodes[segment]:
                out_point_image[point]=[]
        for segment in out_polygon.line_segments:
            projection_points.append([]) 
            
        for segment in in_polygon.line_segments:
           
            segment_to_nodes[segment]=segment.discretise(0)
       
                                
        for out_segment in out_polygon.line_segments:
            
            index_0=out_polygon.line_segments.index(out_segment)
            
            mid=out_segment.point_at_abscissa(out_segment.length()/2)
            
            d=[]
            for in_segment in in_polygon.line_segments:
                  
      
                  l=in_segment.point_distance(mid)
                  d.append(l)
            index=d.index(min(d))
            
               
            near_segment=in_polygon.line_segments[index]
            
            closest_segment[out_segment]=near_segment
            for point in segment_to_nodes[out_segment]:
                # point.plot(ax=ax,color='r')
               
                i=[]
                d_1=[]
                index_point=segment_to_nodes[out_segment].index(point)
                projection, _=near_segment.point_projection(point)
                
                for in_segment in in_polygon.line_segments:
                    
                    if point.point_distance(in_segment.start) < point.point_distance(in_segment.end):
                        
                        d_1.append(point.point_distance(in_segment.start))
                        
                        i.append(0)
                    elif  point.point_distance(in_segment.start) > point.point_distance(in_segment.end):
                        d_1.append(point.point_distance(in_segment.end))
                        i.append(1)
                    elif point.point_distance(in_segment.start) == point.point_distance(in_segment.end):
                         
                          d_1.append(point.point_distance(in_segment.start))
                          i.append(0)
                index_1=d_1.index(min(d_1))
                new_proj=in_polygon.line_segments[index_1].points[i[index_1]]
                
                if len(segment_to_nodes[out_segment])==2  : 
                    
                    if out_point_image[point]==[]:
                        out_point_image[point].append([index_0,new_proj])
                        if new_proj not in projection_points[index_0]:
                          projection_points[index_0].insert(index_point,new_proj)
                    else :
                          if out_point_image[point][0][1] != new_proj :
                            
                            projection_points[out_point_image[point][0][0]].remove(out_point_image[point][0][1])
                            projection_points[out_point_image[point][0][0]].append(new_proj)
                            
                          if new_proj not in projection_points[index_0]:
                              projection_points[index_0].insert(index_point,new_proj)
                  
                                 
                           
                          
                else:                                               
                    if projection is None :
                            
                           
                            if out_point_image[point]==[]:
                              
                                out_point_image[point].append([index_0,new_proj])
                                  
                                if new_proj not in projection_points[index_0]:
                                   
                                    projection_points[index_0].insert(index_point,new_proj)
                                
                            else :
                                 
                                 if out_point_image[point][0][1] != new_proj :
                                      
                                      projection_points[out_point_image[point][0][0]].remove(out_point_image[point][0][1])
                                      projection_points[out_point_image[point][0][0]].append(new_proj)
                                      if new_proj not in projection_points[index_0]:
                                            projection_points[index_0].insert(index_point,new_proj)
                                     
                                 else :
                                     if new_proj not in projection_points[index_0]:
                                        projection_points[index_0].insert(index_point,new_proj)
                    else :
                        x=0
                        if out_point_image[point]==[]:
                          i=0
                          
                          if projection.point_distance(near_segment.start) > projection.point_distance(near_segment.end) :
                             i+=1
                          l = near_segment.length()/projection.point_distance(near_segment.points[i])
                          
                          if out_segment.length()/projection.point_distance(near_segment.points[i])<3:
                             x=10
                          else :
                              x=3
                          if l <= x :
                           
                            out_point_image[point].append([index_0,projection]) 
                            projection_points[index_0].insert(index_point,projection)
                          else :
                               
                               if new_proj not in projection_points[index_0]:     
                                   projection_points[index_0].insert(index_point,new_proj)
                                      
                        else :
                             if new_proj not in projection_points[index_0]:     
                                   projection_points[index_0].insert(index_point,new_proj)
                  
        for out_segment in out_polygon.line_segments:
            
            index_0=out_polygon.line_segments.index(out_segment)
            v=len(segment_to_nodes[out_segment])
            u=len(projection_points[index_0])

            if u>=v and u>2:
              
                for j in range(v-1):
                    
                    new_triangle=volmdlr.wires.Triangle2D([projection_points[index_0][j],projection_points[index_0][j+1],
                                        segment_to_nodes[out_segment][j]])   

                      
          
                    all_triangles.append(new_triangle)
                    new_triangle_0=volmdlr.wires.Triangle2D([projection_points[index_0][j+1],segment_to_nodes[out_segment][j],
                                        segment_to_nodes[out_segment][j+1]])   
                   

                
                    all_triangles.append(new_triangle)
                    new_triangle_0=volmdlr.wires.Triangle2D([projection_points[index_0][j+1],segment_to_nodes[out_segment][j],
                                        segment_to_nodes[out_segment][j+1]])   
                

                    all_triangles.append(new_triangle_0)
                    
                for j in  range(v-1,u-1):
                      new_triangle=volmdlr.wires.Triangle2D([projection_points[index_0][j],projection_points[index_0][j+1],
                                      segment_to_nodes[out_segment][v-1]])   
                      all_triangles.append(new_triangle)
                      
           
            if u<v :
                for j in range(u-1):
                  
                   new_triangle=volmdlr.wires.Triangle2D([projection_points[index_0][j],segment_to_nodes[out_segment][j+1],
                                      segment_to_nodes[out_segment][j]])   
                                
                   all_triangles.append(new_triangle)
                   new_triangle_0=volmdlr.wires.Triangle2D([projection_points[index_0][j+1],projection_points[index_0][j],
                                       segment_to_nodes[out_segment][j+1]])   

                   all_triangles.append(new_triangle_0)
                   
                for j in  range(u-1,v-1):
                   new_triangle=volmdlr.wires.Triangle2D([segment_to_nodes[out_segment][j],  segment_to_nodes[out_segment][j+1],
                                  projection_points[index_0][u-1]])   
                   all_triangles.append(new_triangle)
                   
            if u==2 and v==2:
                if projection_points[index_0][1] != projection_points[index_0][0] :
                    new_triangle_1=volmdlr.wires.Triangle2D([projection_points[index_0][0],projection_points[index_0][1],
                                       segment_to_nodes[out_segment][1]]) 
                    all_triangles.append(new_triangle_1)
                  
                new_triangle_2=volmdlr.wires.Triangle2D([projection_points[index_0][0],segment_to_nodes[out_segment][0],
                                    segment_to_nodes[out_segment][1]]) 
               
                all_triangles.append(new_triangle_2)
           
                
           
                                
  
        if empty is False:
            
            last_points=[]
            for k in range(len(projection_points)):
                for point in projection_points[k]:
                    if point not in last_points:
                        last_points.append(point)
            last_polygon=volmdlr.wires.ClosedPolygon2D(last_points)
            ear=self.earclip(last_polygon)
            for triangle in ear : 
                if triangle.area<10e-9:
                    ear.remove(triangle)
            all_triangles+=ear
        
        return all_triangles
  
    
    
    def polygon_to_triangles(self,polygons:List[volmdlr.wires.ClosedPolygon2D]):
        triangles=[]
        for polygon in polygons:
            triangles.append(volmdlr.wires.Triangle2D(polygon.points))
        return triangles

    
    def triangulation_max_aspect_ratio(self,triangles:List[volmdlr.wires.Triangle2D]):
        all_aspect_ratios=[]
        for triangle in triangles:
           
            if triangle.area<10E-9:
                return 0
            else :
                all_aspect_ratios.append(triangle.aspect_ratio())
        index=all_aspect_ratios.index(max(all_aspect_ratios))
        
        return all_aspect_ratios[index]
    def triangulation_min_aspect_ratio(self,triangles:List[volmdlr.wires.Triangle2D]):
        all_aspect_ratios=[]
        for triangle in triangles:
            all_aspect_ratios.append(triangle.aspect_ratio())
        index=all_aspect_ratios.index(min(all_aspect_ratios))
        
        return all_aspect_ratios[index]
    def generate_mesh(self,min_aspect_ratio:float,split:bool):
        
        ax=plt.subplot()
        segment_to_nodes={}
        all_segments=set()
        triangles=[]
        interior_polygons=[]
        exterior_polygons=[]
        split_polygons=[]
              
        if self.interior_contours :
            
            for contour in self.interior_contours:
                   
                interior_polygons.append(contour.to_polygon(self.nodes_len))  
                

                     
      
      
       
        for contour in self.exterior_contours:
                 
              exterior_polygons.append(contour.to_polygon(self.nodes_len))   
              
    
         
        
        if split is True :
         

            Surface=volmdlr.faces.Surface2D(self.exterior_contours[-1],self.interior_contours)

            split_contours = Surface.split_at_centers()
             
            if self.interior_contours : 
                for contour in split_contours:
                    
                    split_polygons.append(contour.to_polygon(self.nodes_len)) 
          
    
                
            
            for s in split_polygons:
                s.plot()
                triangles+=self.alternative_triangulation(s,None,False,False)
            
        else :
                      
            """
            
            The convention adopted is the following : The last polygon of 
            interior_polygons(which stays empty) is inside the last polygon of 
            exterior_polygon 
            
            """ 
             
         
            if self.interior_contours:
                if exterior_polygons[-1].polygon_distance(interior_polygons[-1]) > exterior_polygons[-1].max_length()/3 :
                    
                    triangles+=self.alternative_triangulation(exterior_polygons[-1],interior_polygons[-1],True,True)
                else :                
                    
                    triangles+=self.mesh_in_between(interior_polygons[-1],exterior_polygons[-1],True)
            
                if len(exterior_polygons)>1:
                    for polygon in exterior_polygons:
                      
                        if polygon != exterior_polygons[-1]:
                            
                            polygon_1=polygon.copy()
                           
                            possible_triangles=self.earclip(polygon)
                            
                            """For one polygon,earclip created a LineSegment2D as an instance of Triangle2D
                            for unknown reasons,hence the area test below """
                            
                            for triangle in possible_triangles:
                                if triangle.area<10E-9:
                                    possible_triangles.remove(triangle)
                                    
                                    
                            if min_aspect_ratio == None:
                              triangles+=possible_triangles
                             
                            else: 
                                Next=True
                                k=0
                                while Next is True and k < len(possible_triangles):
                                 
                                      triangle=possible_triangles[k]    
                                     
                                      if triangle.aspect_ratio() > min_aspect_ratio:
                                        Next=False
                                        triangles+=self.alternative_triangulation(polygon_1,None,False,False)
                                        
                                      else :
                                      
                                          k=k+1
                                          
                                if Next==True :
                                    triangles+=possible_triangles
                          
            else :
                for polygon in exterior_polygons:
                      
                    polygon_1=polygon.copy()
                   
                    possible_triangles=self.earclip(polygon)
                    
                    """For one polygon,earclip created a LineSegment2D as an instance of Triangle2D
                    for unknown reasons,hence the area test below """
                    
                    for triangle in possible_triangles:
                        if triangle.area==0:
                            possible_triangles.remove(triangle)
                            
                            
                    if min_aspect_ratio == None:
                      triangles+=possible_triangles
                     
                    else: 
                        Next=True
                        k=0
                        while Next is True and k < len(possible_triangles):
                         
                              triangle=possible_triangles[k]    
                             
                              if triangle.aspect_ratio() > min_aspect_ratio:
                                Next=False
                                triangles+=self.alternative_triangulation(polygon_1,None,False,False,False)
                                
                              else :
                              
                                  k=k+1
                                  
                        if Next==True :
                            triangles+=possible_triangles

        all_triangles=[]
        all_triangles+=triangles
        all_triangle_elements=[]
        plot_aspect_ratio_triangles=[]
        all_aspect_ratios={}
        
        # for triangle in all_triangles:

        #       triangle.plot(ax=ax)


        # #       triangle.plot(ax=ax)
        #       plot_aspect_ratio_triangles.append(triangle)
        #       all_aspect_ratios.update({triangle:triangle.aspect_ratio()})

        for triangle in triangles:
            
            all_segments= all_segments.union(triangle.line_segments)
            
        for segment in all_segments:
            
            segment_to_nodes[segment]=segment.discretise(self.nodes_len)
                
        for triangle in triangles :
            
            meshing=triangle.mesh_triangle(segment_to_nodes,self.nodes_len)
            plot_aspect_ratio_triangles+=meshing[0]
            all_triangles+=meshing[0]
            all_aspect_ratios.update(meshing[1])

        for triangle in all_triangles : 
            for point in triangle.points:
                point.plot(ax=ax,color='r')
        
                
                
        for triangle in all_triangles:
  
            triangular_element=TriangularElement(triangle.points)
            all_triangle_elements.append(triangular_element)
            

        self.plot_aspect_ratio(plot_aspect_ratio_triangles,all_aspect_ratios,ax)
        ax.set_aspect('equal')
        return all_triangle_elements
    
    def plot_aspect_ratio(self,all_triangles:List[volmdlr.wires.Triangle2D],
                          all_aspect_ratios:Dict[volmdlr.wires.Triangle2D,float],
                          ax,min_aspect_ratio=None,max_aspect_ratio=None):
        
        """
        Plots the mesh with colored triangles representing the \
        value of the aspect ratio. 
        """
        color_map = ((0,0,1), (1,0,0))
        if ax is None :
            fig, ax = plt.subplots()
        else :
            fig = plt.gcf()
        A= [a for a in list(all_aspect_ratios.values())]
        
        if max_aspect_ratio is None:
           max_aspect = max(A)
           
        if min_aspect_ratio is None:
           min_aspect = min(A)
     
        aspect_ratio_to_color = {}
        for a in A:
            if a > max_aspect:
                x = 1
            else:
                x = (a - min_aspect) / (max_aspect - min_aspect)
            color = (color_map[0][0]-(color_map[0][0]-color_map[1][0])*x, 
                     color_map[0][1]-(color_map[0][1]-color_map[1][1])*x,
                     color_map[0][2]-(color_map[0][2]-color_map[1][2])*x)
            aspect_ratio_to_color[a] = color
        
        for triangle in all_triangles:
            triangle.plot(ax=ax, color=aspect_ratio_to_color[all_aspect_ratios[triangle]], 
                          fill=True) 
            
        norm = matplotlib.colors.Normalize(vmin=min_aspect, vmax=max_aspect)
        sm = plt.cm.ScalarMappable(cmap=blue_red, norm=norm) 
        sm.set_array([])
        cbar = fig.colorbar(sm, ticks=npy.linspace(min_aspect,max_aspect, 10))
        cbar.set_label('Aspect Ratio')
        
        return ax    
      
    
        
               
           
           
       
       
            
        
        
    
  
        
                    
            
       
        
    
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 28 11:26:19 2020

@author: masfaraud
"""

import dessia_common.workflow 
import volmdlr.core
import volmdlr

packer = dessia_common.workflow.Sequence(3, float)

point3D_instanciation = dessia_common.workflow.InstanciateModel(
    volmdlr.Point3D)
pipe_vector = dessia_common.workflow.Pipe(packer.outputs[0],
                                          point3D_instanciation.inputs[0])
point3D_instanciator = dessia_common.workflow.Workflow(
    [packer, point3D_instanciation],
    [pipe_vector],
    point3D_instanciation.outputs[0],
    name='Point3D instanciator')
point3D_instanciator.inputs[0].name = 'X'
point3D_instanciator.inputs[1].name = 'Y'
point3D_instanciator.inputs[2].name = 'Z'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
<<<<<<< HEAD
Created on Wed Oct 23 10:55:23 2019
=======
Created on Fri Oct 25 14:53:57 2019
>>>>>>> 45a9a2643a660a78f83893310fdfd943ab74b0a7

@author: ringhausen
"""

import math
import sys
from collections import namedtuple

Point = namedtuple('Point', ['x', 'y'])

EPSILON = math.sqrt(sys.float_info.epsilon)


def earclip(polygon):
    """
    Simple earclipping algorithm for a given polygon p.
    polygon is expected to be an array of 2-tuples of the cartesian points of the polygon
    For a polygon with n points it will return n-2 triangles.
    The triangles are returned as an array of 3-tuples where each item in the tuple is a 2-tuple of the cartesian point.
    e.g
    >>> polygon = [(0,1), (-1, 0), (0, -1), (1, 0)]
    >>> triangles = tripy.earclip(polygon)
    >>> triangles
    [((1, 0), (0, 1), (-1, 0)), ((1, 0), (-1, 0), (0, -1))]
    Implementation Reference:
        - https://www.geometrictools.com/Documentation/TriangulationByEarClipping.pdf
    """
    ear_vertex = []
    triangles = []

    polygon = [Point(*point) for point in polygon]

    if _is_clockwise(polygon):
        polygon.reverse()

    point_count = len(polygon)
    for i in range(point_count):
        prev_index = i - 1
        prev_point = polygon[prev_index]
        point = polygon[i]
        next_index = (i + 1) % point_count
        next_point = polygon[next_index]

        if _is_ear(prev_point, point, next_point, polygon):
            ear_vertex.append(point)

    while ear_vertex and point_count >= 3:
        ear = ear_vertex.pop(0)
        i = polygon.index(ear)
        prev_index = i - 1
        prev_point = polygon[prev_index]
        next_index = (i + 1) % point_count
        next_point = polygon[next_index]

        polygon.remove(ear)
        point_count -= 1
        triangles.append(((prev_point.x, prev_point.y), (ear.x, ear.y), (next_point.x, next_point.y)))

        if point_count > 3:
            prev_prev_point = polygon[prev_index - 1]
            next_next_index = (i + 1) % point_count
            next_next_point = polygon[next_next_index]

            groups = [
                (prev_prev_point, prev_point, next_point, polygon),
                (prev_point, next_point, next_next_point, polygon),
            ]
            for group in groups:
                p = group[1]
                if _is_ear(*group):
                    if p not in ear_vertex:
                        ear_vertex.append(p)
                elif p in ear_vertex:
                    ear_vertex.remove(p)

    return triangles


def _is_clockwise(polygon):
    s = 0
    polygon_count = len(polygon)
    for i in range(polygon_count):
        point = polygon[i]
        point2 = polygon[(i + 1) % polygon_count]
        s += (point2.x - point.x) * (point2.y + point.y)
    return s > 0


def _is_convex(prev, point, next):
    return _triangle_sum(prev.x, prev.y, point.x, point.y, next.x, next.y) < 0


def _is_ear(p1, p2, p3, polygon):
    ear = _contains_no_points(p1, p2, p3, polygon) and \
        _is_convex(p1, p2, p3) and \
        _triangle_area(p1.x, p1.y, p2.x, p2.y, p3.x, p3.y) > 0
    return ear


def _contains_no_points(p1, p2, p3, polygon):
    for pn in polygon:
        if pn in (p1, p2, p3):
            continue
        elif _is_point_inside(pn, p1, p2, p3):
            return False
    return True


def _is_point_inside(p, a, b, c):
    area = _triangle_area(a.x, a.y, b.x, b.y, c.x, c.y)
    area1 = _triangle_area(p.x, p.y, b.x, b.y, c.x, c.y)
    area2 = _triangle_area(p.x, p.y, a.x, a.y, c.x, c.y)
    area3 = _triangle_area(p.x, p.y, a.x, a.y, b.x, b.y)
    areadiff = abs(area - sum([area1, area2, area3])) < EPSILON
    return areadiff


def _triangle_area(x1, y1, x2, y2, x3, y3):
    return abs((x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2)) / 2.0)


def _triangle_sum(x1, y1, x2, y2, x3, y3):
    return x1 * (y3 - y2) + x2 * (y1 - y3) + x3 * (y2 - y1)


def calculate_total_area(triangles):
    result = []
    for triangle in triangles:
        sides = []
        for i in range(3):
            next_index = (i + 1) % 3
            pt = triangle[i]
            pt2 = triangle[next_index]
            # Distance between two points
            side = math.sqrt(math.pow(pt2[0] - pt[0], 2) + math.pow(pt2[1] - pt[1], 2))
            sides.append(side)
        # Heron's numerically stable forumla for area of a triangle:
        # https://en.wikipedia.org/wiki/Heron%27s_formula
        # However, for line-like triangles of zero area this formula can produce an infinitesimally negative value
        # as an input to sqrt() due to the cumulative arithmetic errors inherent to floating point calculations:
        # https://people.eecs.berkeley.edu/~wkahan/Triangle.pdf
        # For this purpose, abs() is used as a reasonable guard against this condition.
        c, b, a = sorted(sides)
        area = .25 * math.sqrt(abs((a + (b + c)) * (c - (a - b)) * (c + (a - b)) * (a + (b - c))))
        result.append((area, a, b, c))
    triangle_area = sum(tri[0] for tri in result)
    return triangle_area


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#cython: language_level=3
"""

Cython functions

"""
# from __future__ import annotations
from typing import TypeVar, List, Tuple
import math
from dessia_common import DessiaObject
import matplotlib.pyplot as plt
from matplotlib.patches import FancyArrow, FancyArrowPatch
import warnings
import random
import numpy as npy
from mpl_toolkits.mplot3d import proj3d

# =============================================================================

cdef (double, double) Csub2D(double u1, double u2,
                             double v1, double v2):
    return (u1-v1, u2-v2)


# =============================================================================

cdef (double, double) Cadd2D(double u1, double u2,
                             double v1, double v2,):
    return (u1+v1, u2+v2)


# =============================================================================

cdef (double, double) Cmul2D(double u1, double u2, double value):
    return (u1*value, u2*value)

#def mul2D(vector, value):
#    return Cmul2D(vector.x, vector.y, value)

# =============================================================================

cdef double CVector2DDot(double u1, double u2,
                         double v1, double v2):
    return u1*v1 + u2*v2

#def Vector2DDot(vector1, vector2):
#    return CVector2DDot(vector1.x, vector1.y,
#                        vector2.x, vector2.y)

# =============================================================================

cdef double CVector2Dnorm(double u1, double u2):
    return (u1*u1 + u2*u2)**0.5

#def Vector2Dnorm(vector):
#    return CVector2Dnorm(vector.x, vector.y)

# =============================================================================

cdef (double, double, double) Csub3D(double u1, double u2, double u3,
                                     double v1, double v2, double v3):
    return (u1-v1, u2-v2, u3-v3)

#def sub3D(vector1, vector2):
#    return Csub3D(vector1.x, vector1.y, vector1.z,
#                  vector2.x, vector2.y, vector2.z)

# =============================================================================

cdef (double, double, double) Cadd3D(double u1, double u2, double u3,
                                     double v1, double v2, double v3):
    return (u1+v1, u2+v2, u3+v3)

#def add3D(vector1, vector2):
#    return Cadd3D(vector1.x, vector1.y, vector1.z,
#                  vector2.x, vector2.y, vector2.z)

# =============================================================================

cdef (double, double, double) Cmul3D(double u1, double u2, double u3,
                                     double value):
    return (u1*value, u2*value, u3*value)

#def mul3D(vector, value):
#    return Cmul3D(vector.x, vector.y, vector.z, value)

# =============================================================================

cdef double CVector3DDot(double u1, double u2, double u3,
                         double v1, double v2, double v3):
    return u1*v1 + u2*v2 + u3*v3

#def Vector3DDot(vector1, vector2):
#    return CVector3DDot(vector1.x, vector1.y, vector1.z,
#                        vector2.x, vector2.y, vector2.z)

# =============================================================================

cdef double CVector3Dnorm(double u1, double u2, double u3):
    return (u1*u1 + u2*u2 + u3*u3)**0.5

#def Vector3Dnorm(vector):
#    return CVector3Dnorm(vector.x, vector.y, vector.z)


# =============================================================================

cdef (double, double, double) CVector3D_cross(double u1, double u2, double u3,
                                               double v1, double v2, double v3):
    return (u2*v3 - u3*v2, u3*v1 - u1*v3, u1*v2 - u2*v1)

#def vector3D_cross(vector1, vector2):
#    return C_vector3D_cross(vector1.x, vector1.y, vector1.z,
#                            vector2.x, vector2.y, vector2.z)


# =============================================================================

cdef (double, double, double) C_vector3D_rotation(double vx, double vy, double vz,
                                                  double center_x, double center_y, double center_z,
                                                  double axis_x, double axis_y, double axis_z,
                                                  double angle):

    cdef double ux = vx - center_x
    cdef double uy = vy - center_y
    cdef double uz = vz - center_z

    cdef double cos_angle = math.cos(angle)
    cdef double sin_angle = math.sin(angle)

    cdef double rv1_x = cos_angle*ux
    cdef double rv1_y = cos_angle*uy
    cdef double rv1_z = cos_angle*uz

    rv2_x, rv2_y, rv2_z = Cmul3D(axis_x, axis_y, axis_z,
                                 (1-cos_angle)*CVector3DDot(
                                         ux, uy, uz,
                                         axis_x, axis_y, axis_z)
                                 )

    rv3_x, rv3_y, rv3_z = CVector3D_cross(axis_x, axis_y, axis_z,
                                           ux, uy, uz)

    return (rv1_x + rv2_x + rv3_x*sin_angle + center_x,
            rv1_y + rv2_y + rv3_y*sin_angle + center_y,
            rv1_z + rv2_z + rv3_z*sin_angle + center_z)

def vector3D_rotation(vector, center, axis, angle):
        return C_vector3D_rotation(vector.x, vector.y, vector.z,
                                   center.x, center.y, center.z,
                                   axis.x, axis.y, axis.z,
                                   angle)



cdef (double, double, double) C_matrix_vector_multiplication3(double M11, double M12, double M13,
                                                              double M21, double M22, double M23,
                                                              double M31, double M32, double M33,
                                                              double v1, double v2, double v3):

    return (M11*v1 + M12*v2 + M13*v3,
            M21*v1 + M22*v2 + M23*v3,
            M31*v1 + M32*v2 + M33*v3)


cdef (double, double, double,
      double, double, double,
      double, double, double) Cmatrix_multiplication3(double A11, double A12, double A13,
                                                      double A21, double A22, double A23,
                                                      double A31, double A32, double A33,
                                                      double B11, double B12, double B13,
                                                      double B21, double B22, double B23,
                                                      double B31, double B32, double B33):

    return (A11*B11 + A12*B21 + A13*B31,
            A11*B12 + A12*B22 + A13*B32,
            A11*B13 + A12*B23 + A13*B33,
            A21*B11 + A22*B21 + A23*B31,
            A21*B12 + A22*B22 + A23*B32,
            A21*B13 + A22*B23 + A23*B33,
            A31*B11 + A32*B21 + A33*B31,
            A31*B12 + A32*B22 + A33*B32,
            A31*B13 + A32*B23 + A33*B33)


# =============================================================================

def polygon_point_belongs(point, points):

    cdef int i
    cdef int n = len(points)
    cdef bint inside = False
    cdef float x, y, p1x, p1y, p2x, p2y, xints
    x,y=point
    p1x,p1y = points[0]

    for i in range(n+1):
        p2x, p2y = points[i % n]
        if y > min(p1y,p2y):
            if y <= max(p1y,p2y):
                if x <= max(p1x,p2x):
                    if p1y != p2y:
                        xints = (y-p1y)*(p2x-p1x)/(p2y-p1y)+p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y

    return inside

# =============================================================================

cdef (double, (double, double)) CLineSegment2DPointDistance((double, double) p1, (double, double) p2, (double, double) point):
    cdef double t
    cdef (double, double) u, projection

    u = (p2[0] - p1[0], p2[1] - p1[1])
    ppx, ppy = Csub2D(point[0], point[1], p1[0], p1[1])
    t = max(0, min(1, CVector2DDot(ppx, ppy, u[0], u[1]) / CVector2Dnorm(u[0], u[1])**2))
    vx, vy  = Cmul2D(u[0], u[1], t)
    projection = Cadd2D(p1[0], p1[1], vx, vy)
    ppx, ppy = projection[0]-point[0], projection[1]-point[1]
    return CVector2Dnorm(ppx, ppy), projection

def LineSegment2DPointDistance(points, point):
    return CLineSegment2DPointDistance(tuple(points[0]), tuple(points[1]), tuple(point))


# =============================================================================
#  Points, Vectors
# =============================================================================

class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def plot2d(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        FancyArrowPatch.draw(self, renderer)

    def plot(self, ax=None, color='b'):
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

        points = [self.start, self.end]
        x = [p.x for p in points]
        y = [p.y for p in points]
        z = [p.z for p in points]
        ax.plot(x, y, z, 'o-k')
        return ax

class Vector(DessiaObject):
    """
    Abstract class of vector
    """
    def __radd__(self, other_vector):
        return self + other_vector

    def __rsub__(self, other_vector):
        return self - other_vector

    def __rmul__(self, value):
        return self * value

    def __rtruediv__(self, value):
        return self / value

    def __lt__(self, other_vector):
        return self.norm() < other_vector.norm()

    def __le__(self, other_vector):
        return self.norm() <= other_vector.norm()


    def is_colinear_to(self, other_vector):
        try:
            return math.isclose(abs(self.dot(other_vector))/self.norm()/other_vector.norm(),
                                1,
                                abs_tol=1e-5)
            
        except ZeroDivisionError:
            return False

    @classmethod
    def mean_point(cls, points):
        n = 1
        point = points[0].copy()
        for point2 in points[1:]:
            point += point2
            n += 1
        point /= n
        return point

class Vector2D(Vector):
    def __init__(self, x:float, y:float, name=''):
        self.x = x
        self.y = y
        self.name = name

    def __repr__(self):
        return '{}: [{}, {}]'.format(self.__class__.__name__, self.x, self.y)

    def __setitem__(self, key, item):
        if key == 0:
            self.x = item
        elif key == 1:
            self.y = item
        else:
            raise IndexError

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        else:
            raise IndexError

    def __add__(self, other_vector):
        return Vector2D(*Cadd2D(self.x, self.y,
                               other_vector.x, other_vector.y))

    def __neg__(self):
        return Vector2D(-self.x, -self.y)

    def __sub__(self, other_vector):
        return Vector2D(*Csub2D(self.x, self.y,
                                other_vector.x, other_vector.y))

    def __mul__(self, value:float):
        return Vector2D(*Cmul2D(self.x, self.y, value))

    def __truediv__(self, value:float):
        if value == 0:
            raise ZeroDivisionError
        return Vector2D(self.x / value,
                        self.y / value)

    def __round__(self, ndigits:int=6):
        return self.__class__(round(self.x, ndigits),
                              round(self.y, ndigits))

    def __hash__(self):
        return 0


    def __eq__(self, other_vector):
        if other_vector.__class__.__name__ not in ['Vector2D', 'Point2D']:
            return False
        return math.isclose(self.x, other_vector.x, abs_tol=1e-06) \
        and math.isclose(self.y, other_vector.y, abs_tol=1e-06)

    def norm(self):
        """
        :returns: norm of vector
        """
        return CVector2Dnorm(self.x, self.y)

    def normalize(self):
        """
        normalize the vector modifying its coordinates in place
        """
        n = self.norm()
        if math.isclose(n, 0, abs_tol=1e-9):
            raise ZeroDivisionError

        self.x /= n
        self.y /= n

    def dot(self, other_vector):
        return CVector2DDot(self.x,
                            self.y,
                            other_vector.x,
                            other_vector.y)

    def cross(self, other_vector):
        return self.x*other_vector.y - self.y*other_vector.x

    def point_distance(self, point2):
        return (self-point2).norm()


    def rotation(self, center, angle, copy=True):
        u = self - center
        v2x = math.cos(angle)*u[0] - math.sin(angle)*u[1] + center[0]
        v2y = math.sin(angle)*u[0] + math.cos(angle)*u[1] + center[1]
        if copy:
            return self.__class__(v2x, v2y)
        else:
            self.x = v2x
            self.y = v2y

    def translation(self, offset, copy=True):
        """
        :param offset: an other Vector2D
        """
        v2x = self.x + offset[0]
        v2y = self.y + offset[1]
        if copy:
            return self.__class__(v2x, v2y)
        else:
            self.x = v2x
            self.y = v2y

    def frame_mapping(self, frame, side, copy=True):
        """
        side = 'old' or 'new'
        """
        if side == 'old':
            new_vector = frame.old_coordinates(self)
            if copy:
                return new_vector
            else:
                self.x = new_vector.x
                self.y = new_vector.y

        if side == 'new':
            new_vector = frame.new_coordinates(self)
            if copy:
                return new_vector
            else:
                self.x = new_vector.x
                self.y = new_vector.y


    def to_3d(self, plane_origin, vx, vy):
        return Vector3D(plane_origin.x + vx.x*self.x + vy.x*self.y,
                        plane_origin.y + vx.y*self.x + vy.y*self.y,
                        plane_origin.z + vx.z*self.x + vy.z*self.y,
                        )

    def to_point(self):
        return Point2D(self.x, self.y)

    def normal_vector(self):
        n = Vector2D(-self.y, self.x)
        return n

    def unit_normal_vector(self):
        n = self.normal_vector()
        n.normalize()
        return n


    def deterministic_unit_normal_vector(self):
        return self.unit_normal_vector()

    @classmethod
    def random(cls, xmin, xmax, ymin, ymax):
        return cls(random.uniform(xmin, xmax),
                   random.uniform(ymin, ymax))


    def plot(self, amplitude=0.5, width=None, head_width=None, origin=None,
             ax=None, color='k', line=False, label=None, normalize=False):
        if origin is None:
            origin = Vector2D(0., 0.)

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        if math.isclose(self.norm(), 0, abs_tol=1e-9):
            point = origin.copy()
            point.plot(ax=ax, color=color)
            return ax

        if width is None:
            width = 0.001*5*amplitude
        if head_width is None:
            head_width = 0.3*amplitude

        if not normalize:
            ax.add_patch(FancyArrow(origin[0], origin[1],
                                    self.x*amplitude, self.y*amplitude,
                                    width=width,
                                    head_width=head_width,
                                    length_includes_head=True,
                                    color=color))
        else:
            normalized_vector = self.copy()
            normalized_vector.normalize()
            ax.add_patch(FancyArrow(origin[0], origin[1],
                                    normalized_vector.x*amplitude,
                                    normalized_vector.y*amplitude,
                                    width=width,
                                    head_width=head_width,
                                    length_includes_head=True,
                                    color=color))

        if line:
            style='-'+color
            linestyle = '-.'
            origin = Point2D(origin)
            p1, p2 = origin, origin+self
            u = p2 - p1
            p3 = p1 - 3*u
            p4 = p2 + 4*u
            ax.plot([p3[0], p4[0]], [p3[1], p4[1]], style, linestyle=linestyle)

        if label is not None:
            ax.text(*(origin+self*amplitude), label)

        return ax


X2D = Vector2D(1, 0)
Y2D = Vector2D(0, 1)


class Point2D(Vector2D):
    def __init__(self, x:float, y: float, name: str=''):
        Vector2D.__init__(self, x=x, y=y, name=name)

    def __add__(self, other_vector):
        return Point2D(*Cadd2D(self.x, self.y, other_vector.x, other_vector.y))

    def __neg__(self):
        return Point2D(-self.x, -self.y)

    def __sub__(self, other_vector):
        return Point2D(*Csub2D(self.x, self.y,
                               other_vector.x, other_vector.y))

    def __mul__(self, value:float):
        return Point2D(*Cmul2D(self.x, self.y, value))

    def __truediv__(self, value:float):
        if value == 0:
            raise ZeroDivisionError
        return Point2D(self.x / value,
                       self.y / value)

    def to_3d(self, plane_origin, vx, vy):
        return Point3D(plane_origin.x + vx.x*self.x + vy.x*self.y,
                       plane_origin.y + vx.y*self.x + vy.y*self.y,
                       plane_origin.z + vx.z*self.x + vy.z*self.y)

    def to_vector(self):
        return Vector2D(self.x, self.y, self.z)

    def plot(self, ax=None, color='k', plot_points=True):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot([self.x], [self.y], color=color, marker='o')
        return ax

    def point_distance(self, point2:'Point2D'):
        return (self-point2).norm()


    @classmethod
    def line_intersection(cls, line1, line2, curvilinear_abscissa=False):
        x1 = line1.points[0][0]
        y1 = line1.points[0][1]
        x2 = line1.points[1][0]
        y2 = line1.points[1][1]
        x3 = line2.points[0][0]
        y3 = line2.points[0][1]
        x4 = line2.points[1][0]
        y4 = line2.points[1][1]


        denominateur = (x1-x2)*(y3-y4)-(y1-y2)*(x3-x4)
        if math.isclose(denominateur, 0, abs_tol=1e-6):
            if not curvilinear_abscissa:
                return None
            else:
                return None, None, None
        else:
            x = (x1*y2-y1*x2)*(x3-x4)-(x1-x2)*(x3*y4-y3*x4)
            x = x / denominateur
            y = (x1*y2-y1*x2)*(y3-y4)-(y1-y2)*(x3*y4-y3*x4)
            y = y / denominateur
            if not curvilinear_abscissa:
                return cls(x,y)
            else:
                t = (x1-x3)*(y3-y4)-(y1-y3)*(x3-x4)
                t = t / denominateur
                u = (x1-x2)*(y1-y3)-(y1-y2)*(x1-x3)
                u = -u / denominateur
                return (cls(x,y), t, u)

    @classmethod
    def segment_intersection(cls, segment1, segment2,
                             curvilinear_abscissa=False):
        x1 = segment1.points[0].x
        y1 = segment1.points[0].y
        x2 = segment1.points[1].x
        y2 = segment1.points[1].y
        x3 = segment2.points[0].x
        y3 = segment2.points[0].y
        x4 = segment2.points[1].x
        y4 = segment2.points[1].y

        denominateur = (x1-x2)*(y3-y4)-(y1-y2)*(x3-x4)
        if math.isclose(denominateur, 0, abs_tol=1e-6):
            if not curvilinear_abscissa:
                return None
            else:
                return None, None, None

        t = (x1-x3)*(y3-y4)-(y1-y3)*(x3-x4)
        t = t / denominateur
        u = (x1-x2)*(y1-y3)-(y1-y2)*(x1-x3)
        u = -u / denominateur
        if (0 <= t and t <= 1) or (0 <= u and u <= 1):
            x = (x1*y2-y1*x2)*(x3-x4)-(x1-x2)*(x3*y4-y3*x4)
            x = x / denominateur
            y = (x1*y2-y1*x2)*(y3-y4)-(y1-y2)*(x3*y4-y3*x4)
            y = y / denominateur
            if not curvilinear_abscissa:
                return cls((x,y))
            else:
                return (cls((x,y)), t, u)
        else:
            if not curvilinear_abscissa:
                return None
            else:
                return None, None, None

    def plot_data(self, marker=None, color='black', size=1,
                  opacity=1, arrow=False, stroke_width=None):
        return {'type' : 'point',
                'data' : [self.x, self.y],
                'color' : color,
                'marker' : marker,
                'size' : size,
                'opacity' : opacity
                }

    @classmethod
    def middle_point(cls, point1, point2):
        return (point1 + point2)*0.5

    @classmethod
    def line_projection(cls, point, line):
        p1, p2 = line[0], line[1]
        n = line.unit_normal_vector()
        pp1 = point - p1
        return  pp1 - pp1.dot(n)*n + p1

O2D = Point2D(0, 0)

class Vector3D(Vector):

    def __init__(self, x:float, y:float, z:float, name:str=''):
        self.x = x
        self.y = y
        self.z = z
        self.name = name

    def __repr__(self):
        return '{}: [{}, {}, {}]'.format(self.__class__.__name__, self.x, self.y, self.z)

    def __setitem__(self, key, item):
        if key == 0:
            self.x = item
        elif key == 1:
            self.y = item
        elif key == 2:
            self.z = item
        else:
            raise IndexError

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        else:
            raise IndexError

    def __add__(self, other_vector):
        return Vector3D(*Cadd3D(self.x, self.y, self.z,
                                other_vector.x,
                                other_vector.y,
                                other_vector.z))

    def __neg__(self):
        return Vector3D(-self.x, -self.y, -self.z)

    def __sub__(self, other_vector):
        return Vector3D(*Csub3D(self.x, self.y, self.z,
                                other_vector.x,
                                other_vector.y,
                                other_vector.z))

    def __mul__(self, value):
        return Vector3D(*Cmul3D(self.x, self.y, self.z, value))

    def __truediv__(self, value):
        if value == 0:
            raise ZeroDivisionError
        return Vector3D(self.x / value,
                        self.y / value,
                        self.z / value)

    def __round__(self, ndigits:int=6):
        return self.__class__(round(self.x, ndigits),
                              round(self.y, ndigits),
                              round(self.z, ndigits))

    def __hash__(self):
        return 0

    def __eq__(self, other_vector:'Vector3D'):
        if other_vector.__class__.__name__ not in ['Vector3D', 'Point3D']:
            return False
        return math.isclose(self.x, other_vector.x, abs_tol=1e-06) \
        and math.isclose(self.y, other_vector.y, abs_tol=1e-06) \
        and math.isclose(self.z, other_vector.z, abs_tol=1e-06)

    def dot(self, other_vector):
        return CVector3DDot(self.x, self.y, self.z,
                            other_vector.x, other_vector.y, other_vector.z)

    def cross(self, other_vector:'Vector3D') -> 'Vector3D':
        return self.__class__(*CVector3D_cross(self.x, self.y, self.z,
                                              other_vector.x,
                                              other_vector.y,
                                              other_vector.z))

    def norm(self) -> float:
        return CVector3Dnorm(self.x, self.y, self.z)


    def normalize(self) -> None:
        """
        normalize the vector modifying its coordinates
        """
        n = self.norm()
        if n == 0:
            raise ZeroDivisionError

        self.x /= n
        self.y /= n
        self.z /= n

    def point_distance(self, point2:'Vector3D') -> float:
        return (self-point2).norm()

    def rotation(self, center:'Point3D', axis:'Vector3D', angle:float, copy:bool=True):
        """
        rotation of angle around axis.
        Used Rodrigues Formula:
            https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
        """
#        u = self - center
#        vector2 = (math.cos(angle)*u
#                   + (1-math.cos(angle))*(u.Dot(axis))*axis
#                   + math.sin(angle)*axis.cross(u)
#                   + center)
        vector2 = vector3D_rotation(self, center, axis, angle)

        if copy:
            return self.__class__(*vector2)
            # return Point3D(vector2)
        else:
            self.x = vector2[0]
            self.y = vector2[1]
            self.z = vector2[2]

    def x_rotation(self, angle:float, copy:bool=True):
        """
        rotation of angle around X axis.
        """
        cos_angle = math.cos(angle)
        sin_angle = math.sin(angle)

        y1 = cos_angle * self.y + sin_angle * self.z
        z1 = -sin_angle * self.y + cos_angle * self.z


        if copy:
            return Point3D(self.x, y1, z1)
        else:
            self.y = y1
            self.z = z1

    def y_rotation(self, angle:float, copy:bool=True):
        """
        rotation of angle around Y axis.
        """
        cos_angle = math.cos(angle)
        sin_angle = math.sin(angle)

        z1 = cos_angle * self.z + sin_angle * self.x
        x1 = -sin_angle * self.z + cos_angle * self.x


        if copy:
            return Point3D(x1, self.y, z1)
        else:
            self.x = x1
            self.z = z1

    def z_rotation(self, angle:float, copy:bool=True):
        """
        rotation of angle around Z axis.
        """
        cos_angle = math.cos(angle)
        sin_angle = math.sin(angle)

        x1 = cos_angle * self.x + sin_angle * self.y
        y1 = -sin_angle * self.x + cos_angle * self.y


        if copy:
            return Point3D(x1, y1, self.z)
        else:
            self.x = x1
            self.y = y1

    def translation(self, offset, copy=True):
        if copy:
            return self + offset
        else:
            self.x += offset[0]
            self.y += offset[1]
            self.z += offset[2]

    def frame_mapping(self, frame, side, copy=True):
        """
        side = 'old' or 'new'
        """
        if side == 'old':
            new_vector = frame.old_coordinates(self)
            if copy:
                return new_vector
            else:
                self.x = new_vector.x
                self.y = new_vector.y
                self.z = new_vector.z


        if side == 'new':
            new_vector = frame.new_coordinates(self)
            if copy:
                return new_vector
            else:
                self.x = new_vector.x
                self.y = new_vector.y
                self.z = new_vector.z

    def plane_projection3d(self, plane_origin, x, y):
        z = x.cross(y)
        z.normalize()
        return self - z.dot(self-plane_origin)*z

    def plane_projection2d(self, plane_origin, x, y):
        z = x.cross(y)
        z.normalize()
        p3d = self - (self-plane_origin).dot(z)*z
        u1 = p3d.dot(x)
        u2 = p3d.dot(y)
        return Point2D(u1, u2)


    def to_2d(self, plane_origin, x, y):
        x2d = self.dot(x) - plane_origin.dot(x)
        y2d = self.dot(y) - plane_origin.dot(y)
        return Point2D(x2d, y2d)

    def random_unit_normal_vector(self):
        """
        Returns a random normal vector
        """
        v = Vector3D.random(0, 1, 0, 1, 0, 1)

        v = v - v.dot(self)*self/(self.norm()**2)
        v.normalize()
        return v

    def deterministic_unit_normal_vector(self):
        """
        Returns a deterministic normal vector
        """
        v = X3D
        if not math.isclose(self.y, 0, abs_tol=1e-7) \
        or not math.isclose(self.z, 0, abs_tol=1e-7):
            v = X3D
        else:
            v = Y3D
        v = v - v.dot(self)*self/(self.norm()**2)
        v.normalize()
        return v

    def copy(self):
        return Vector3D(self.x, self.y, self.z)

    @classmethod
    def random(cls, xmin, xmax, ymin, ymax, zmin, zmax):
        return cls(random.uniform(xmin, xmax),
                   random.uniform(ymin, ymax),
                   random.uniform(zmin, zmax))

    def to_point(self):
        return Point3D(self.x, self.y, self.z)

    @classmethod
    def from_step(cls, arguments, object_dict):
        if type(arguments[1]) is int:
        # VECTOR
            return cls(*object_dict[arguments[1]], arguments[0][1:-1])
        else:
        # DIRECTION
            # return cls(*[float(i)/1000 for i in arguments[1][1:-1].split(",")],
            #             arguments[0][1:-1])
            return cls(*[float(i) for i in arguments[1][1:-1].split(",")],
                        arguments[0][1:-1])

    def to_step(self, current_id, vector=False):
        content = "#{} = DIRECTION('{}',({},{},{}));\n"\
                        .format(current_id, self.name,
                                round(self.x, 6),
                                round(self.y, 6),
                                round(self.z, 6))
        if vector:
            content += "#{} = VECTOR('{}',#{},1.);\n".format(current_id+1,
                                                           self.name,
                                                           current_id)
            current_id += 1
        return content, current_id

    def plot(self, ax=None, starting_point=None, color=''):
        if starting_point is None:
            starting_point = Point3D(0, 0, 0)
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        xs = [starting_point[0], self.x+starting_point[0]]
        ys = [starting_point[1], self.y+starting_point[1]]
        zs = [starting_point[2], self.z+starting_point[2]]
        if color:
            a = Arrow3D(xs, ys, zs, mutation_scale=10, lw=3, arrowstyle="-|>", color=color)
        else:
            a = Arrow3D(xs, ys, zs, mutation_scale=10, lw=3, arrowstyle="-|>")
        ax.add_artist(a)
        return ax


X3D = Vector3D(1, 0, 0)
Y3D = Vector3D(0, 1, 0)
Z3D = Vector3D(0, 0, 1)


class Point3D(Vector3D):
    _standalone_in_db = False

    def __init__(self, x: float, y: float, z: float, name:str=''):
        Vector3D.__init__(self, x, y, z, name)

    def __add__(self, other_vector):
        return Point3D(*Cadd3D(self.x, self.y, self.z,
                              other_vector.x,
                              other_vector.y,
                              other_vector.z))

    def __neg__(self):
        return Point3D(-self.x, -self.y, -self.z)

    def __sub__(self, other_vector):
        return Point3D(*Csub3D(self.x, self.y, self.z,
                               other_vector.x, other_vector.y, other_vector.z))

    def __mul__(self, value):
        return Point3D(*Cmul3D(self.x, self.y, self.z, value))

    def __truediv__(self, value):
        if value == 0:
            raise ZeroDivisionError
        return Point3D(self.x / value,
                       self.y / value,
                       self.z / value)

    def copy(self):
        return Point3D(self.x, self.y, self.z)


    def plot(self, ax=None, color='k', alpha=1):

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

        ax.plot([self.x], [self.y], [self.z], color=color, alpha=alpha,
                marker='o')
        return ax


    def to_2d(self, plane_origin, x, y):
        x2d = self.dot(x) - plane_origin.dot(x)
        y2d = self.dot(y) - plane_origin.dot(y)
        return Point2D(x2d,y2d)

    @classmethod
    def from_step(cls, arguments, object_dict):
        return cls(*[float(i)/1000 for i in arguments[1][1:-1].split(",")],
                    arguments[0][1:-1])

    def to_vector(self):
        return Vector3D(self.x, self.y, self.z)

    def to_step(self, current_id, vertex=False):
        content = "#{} = CARTESIAN_POINT('{}',({},{},{}));\n"\
                        .format(current_id, self.name,
                                round(1000.*self.x, 6),
                                round(1000.*self.y, 6),
                                round(1000.*self.z, 6))
        if vertex:
            content += "#{} = VERTEX_POINT('{}',#{});\n".format(current_id+1,
                                                                self.name,
                                                                current_id)
            current_id += 1

        return content, current_id


    def babylon_script(self):
        s = 'var sphere = BABYLON.MeshBuilder.CreateSphere("point", {diameter: 0.05}, scene);\n'
        s += "sphere.setPositionWithLocalVector(new BABYLON.Vector3({},{},{}));\n".format(self.x,self.y,self.z)
        s += 'var mat = new BABYLON.StandardMaterial("mat", scene);\n'
        s += 'mat.diffuseColor = new BABYLON.Color3(1, 0, 0);\n'
        s += 'sphere.material = mat;\n'
        return s

O3D = Point3D(0, 0, 0)

# =============================================================================
#  Basis, Frames
# =============================================================================

class Matrix22:
    def __init__(self, M11:float, M12:float, M21:float, M22:float):
        self.M11 = M11
        self.M12 = M12
        self.M21 = M21
        self.M22 = M22

    def __add__(self, other_matrix):
        return Matrix22(self.M11 + other_matrix.M11,
                        self.M12 + other_matrix.M12,
                        self.M21 + other_matrix.M21,
                        self.M22 + other_matrix.M22,
                        )

    def __mul__(self, other_matrix):
        return Matrix22(self.M11*other_matrix.M11 + self.M12*other_matrix.M21,
                        self.M11*other_matrix.M12 + self.M12*other_matrix.M22,
                        self.M21*other_matrix.M11 + self.M22*other_matrix.M21,
                        self.M21*other_matrix.M12 + self.M22*other_matrix.M22)

    def determinent(self):
        return self.M11 * self.M22 - self.M12 * self.M21

    def inverse(self):
        det = self.determinent()
        if not math.isclose(det, 0, abs_tol=1e-10):
            det_inv = 1/self.determinent()
            return Matrix22(det_inv*self.M22, -det_inv*self.M12,
                            -det_inv*self.M21, det_inv*self.M11)
        else:
            raise ValueError('The matrix is singular')

    def vector_multiplication(self, vector):
        return vector.__class__(self.M11*vector.x + self.M12*vector.y,
                                self.M21*vector.x + self.M22*vector.y)


class Matrix33:
    def __init__(self, M11:float, M12:float, M13:float,
                       M21:float, M22:float, M23:float,
                       M31:float, M32:float, M33:float):
        self.M11 = M11
        self.M12 = M12
        self.M13 = M13
        self.M21 = M21
        self.M22 = M22
        self.M23 = M23
        self.M31 = M31
        self.M32 = M32
        self.M33 = M33


    def __add__(self, other_matrix):
        return Matrix33(self.M11 + other_matrix.M11,
                        self.M12 + other_matrix.M12,
                        self.M13 + other_matrix.M13,
                        self.M21 + other_matrix.M21,
                        self.M22 + other_matrix.M22,
                        self.M23 + other_matrix.M23,
                        self.M31 + other_matrix.M31,
                        self.M32 + other_matrix.M32,
                        self.M33 + other_matrix.M33)

    def __mul__(self, other_matrix):
        M11, M12, M13, M21, M22, M23, M31, M32, M33 = Cmatrix_multiplication3(self.M11, self.M12, self.M13,
                                                                               self.M21, self.M22, self.M23,
                                                                               self.M31, self.M32, self.M33,
                                                                               other_matrix.M11, other_matrix.M12, other_matrix.M13,
                                                                               other_matrix.M21, other_matrix.M22, other_matrix.M23,
                                                                               other_matrix.M31, other_matrix.M32, other_matrix.M33)

        return Matrix33(M11, M12, M13, M21, M22, M23, M31, M32, M33)


    def __repr__(self):
        s = '[{} {} {}]\n[{} {} {}]\n[{} {} {}]\n'.format(self.M11, self.M12, self.M13,
                                                          self.M21, self.M22, self.M23,
                                                          self.M31, self.M32, self.M33)
        return s

    def float_multiplication(self, float_value):
        return Matrix33(self.M11*float_value, self.M12*float_value, self.M13*float_value,
                        self.M21*float_value, self.M22*float_value, self.M23*float_value,
                        self.M31*float_value, self.M32*float_value, self.M33*float_value)


    def vector_multiplication(self, vector):
        u1, u2, u3 = C_matrix_vector_multiplication3(self.M11, self.M12, self.M13,
                                                     self.M21, self.M22, self.M23,
                                                     self.M31, self.M32, self.M33,
                                                     vector.x, vector.y, vector.z)

        return vector.__class__(u1, u2, u3)

    def determinent(self):
        det = self.M11*self.M22*self.M33 + self.M12*self.M23*self.M31 \
            + self.M13*self.M21*self.M32 - self.M13*self.M22*self.M31 \
            - self.M23*self.M32*self.M11 - self.M33*self.M12*self.M21
        return det

    def inverse(self):
        det = self.determinent()

        if not math.isclose(det, 0, abs_tol=1e-10):
            det_inv = 1/det
            return Matrix33(det_inv*(self.M22*self.M33 - self.M23*self.M32),# a22a33−a23a32
                            det_inv*(self.M13*self.M32 - self.M12*self.M33),# a13a32−a12a33
                            det_inv*(self.M12*self.M23 - self.M13*self.M22),# a12a23−a13a22
                            det_inv*(self.M23*self.M31 - self.M21*self.M33),# a23a31−a21a33
                            det_inv*(self.M11*self.M33 - self.M13*self.M31),# a11a33−a31a13
                            det_inv*(self.M21*self.M13 - self.M23*self.M11),# a13a21−a23a11
                            det_inv*(self.M21*self.M32 - self.M31*self.M22),# a21a32−a31a22
                            det_inv*(self.M12*self.M31 - self.M32*self.M11),# a12a31−a32a11
                            det_inv*(self.M11*self.M22 - self.M21*self.M12) # a11a22−a21a12
                            )
        else:
            raise ValueError('The matrix is singular')

    @classmethod
    def random_matrix(cls, minimum=0, maximum=1):
        range_ = maximum - minimum
        return cls(*[minimum + range_*random.random() for _ in range(9)])

    def to_numpy(self):
        return npy.array([[self.M11, self.M12, self.M13],
                          [self.M21, self.M22, self.M23],
                          [self.M31, self.M32, self.M33]])

class Basis(DessiaObject):
    """
    Abstract class of a basis
    """

    def __contains__(self, vector):
        return vector in self.vectors

    def __hash__(self):
        return hash(self.vectors)

    def copy(self):
        return self.__class__(*self.vectors)



class Basis2D(Basis):
    """
    Defines a 2D basis
    :param u:Vector2D: first vector of the basis
    :param v:Vector2D: second vector of the basis
    """
    def __init__(self, u:Vector2D, v:Vector2D, name:str=''):
        self.u = u
        self.v = v
        self.name = name

    def __eq__(self, other_basis):
        if other_basis.__class__.__name__ != self.__class__.__name__:
            return False
        all_equal = all([other_vector == vector\
                         for other_vector, vector\
                         in zip([other_basis.u, other_basis.v], [self.u, self.v])])
        return all_equal

    def __neg__(self):
        Pinv = self.inverse_transfer_matrix()
        return Basis2D(Vector3D(Pinv[:, 0]),
                       Vector3D(Pinv[:, 1]))

    def __repr__(self):
        return '{}: U={}, V={}'.format(self.__class__.__name__, *self.vectors)

    def _get_vectors(self):
        return (self.u, self.v)

    vectors = property(_get_vectors)


    def to_frame(self, origin:Point3D) -> 'Frame3D':
        return Frame2D(origin, self.u, self.v)


    def transfer_matrix(self):
        return npy.array([[self.u[0], self.v[0]],
                          [self.u[1], self.v[1]]])

    def inverse_transfer_matrix(self):
        det = self.u[0]*self.v[1] - self.v[0]*self.u[1]
        if not math.isclose(det, 0, abs_tol=1e-10):
            return 1/det * npy.array([[self.v[1], -self.v[0]],
                                     [-self.u[1], self.u[0]]])
        else:
            raise ZeroDivisionError

    def new_coordinates(self, vector):
        matrix = self.inverse_transfer_matrix()
        return Point2D((matrix[0][0]*vector.x + matrix[0][1]*vector.y,
                         matrix[1][0]*vector.x + matrix[1][1]*vector.y))


    def old_coordinates(self, vector):
        matrix = self.transfer_matrix()
        return Point2D(matrix[0][0]*vector.x + matrix[0][1]*vector.y,
                       matrix[1][0]*vector.x + matrix[1][1]*vector.y)

    def rotation(self, angle:float, copy=True):
        center = O2D
        new_u = self.u.rotation(center, angle, True)
        new_v = self.v.rotation(center, angle, True)

        if copy:
            return Basis2D(new_u, new_v)
        self.u = new_u
        self.v = new_v

    def copy(self):
        return Basis2D(self.u, self.v)

    def normalize(self):
        """
        normalize the basis modifying its coordinates in place
        """
        self.u.normalize()
        self.v.normalize()


XY = Basis2D(X2D, Y2D)

class Basis3D(Basis):
    """
    Defines a 3D basis
    
    :param u:Vector3D: first vector of the basis
    :param v:Vector3D: second vector of the basis
    :param w:Vector3D: third vector of the basis
    """
    _standalone_in_db = False

    # TODO: create a Basis and Frame class to mutualize between 2D and 2D
    def __init__(self, u:Vector3D, v:Vector3D, w:Vector3D, name:str=''):
        self.u = u
        self.v = v
        self.w = w
        self.name = name

    def __eq__(self, other_basis):
        if other_basis.__class__.__name__ != self.__class__.__name__:
            return False
        all_equal = all([other_vector == vector\
                         for other_vector, vector\
                         in zip([other_basis.u, other_basis.v, other_basis.w], [self.u, self.v, self.w])])
        return all_equal

    def __hash__(self):
        return hash(self.u) + hash(self.v) + hash(self.w)

    def __add__(self, other_basis):
        M = self.transfer_matrix()*other_basis.transfer_matrix()
        return Basis3D(Vector3D(M.M11, M.M21, M.M31),
                       Vector3D(M.M12, M.M22, M.M32),
                       Vector3D(M.M13, M.M23, M.M33))


    def __neg__(self):
        M = self.inverse_transfer_matrix()
        return Basis3D(Vector3D(M.M11, M.M21, M.M31),
                       Vector3D(M.M12, M.M22, M.M32),
                       Vector3D(M.M13, M.M23, M.M33))


    def __sub__(self, other_frame):
        P1inv = other_frame.inverse_transfer_matrix()
        P2 = self.transfer_matrix()
        M = P1inv * P2
        return Basis3D(Vector3D(M.M11, M.M21, M.M31),
                       Vector3D(M.M12, M.M22, M.M32),
                       Vector3D(M.M13, M.M23, M.M33))

    def __round__(self, ndigits:int=6):
        return self.__class__((round(self.u, ndigits),
                               round(self.v, ndigits),
                               round(self.w, ndigits)))

    def __repr__(self):
        return '{}: U={}, V={}, W={}'.format(self.__class__.__name__, *self.vectors)

    def _get_vectors(self):
        return (self.u, self.v, self.w)

    vectors = property(_get_vectors)

    # TODO: transform to annotation when available
    @classmethod
    def from_two_vectors(cls, vector1:Vector3D, vector2:Vector3D) -> 'Basis3D':
        """
        Create a basis with first vector1 adimensionned, as u, v is the vector2 substracted of u component,
        w is the cross product of u and v
        """
        u = vector1.copy()
        u.normalize()
        v = vector2 - vector2.dot(vector1)*vector1
        v.normalize()
        w = u.cross(v)

        return Basis3D(u, v, w)

    def to_frame(self, origin):
        return Frame3D(origin, self.u, self.v, self.w)

    def rotation(self, axis:Vector3D, angle:float, copy:bool=True):
        center = O3D
        new_u = self.u.rotation(center, axis, angle, True)
        new_v = self.v.rotation(center, axis, angle, True)
        new_w = self.w.rotation(center, axis, angle, True)

        if copy:
            return Basis3D(new_u, new_v, new_w, self.name)
        else:
            self.u = new_u
            self.v = new_v
            self.w = new_w

    def x_rotation(self, angle:float, copy:bool=True):
        new_u = self.u.x_rotation(angle, True)
        new_v = self.v.x_rotation(angle, True)
        new_w = self.w.x_rotation(angle, True)

        if copy:
            return Basis3D(new_u, new_v, new_w, self.name)
        else:
            self.u = new_u
            self.v = new_v
            self.w = new_w

    def y_rotation(self, angle:float, copy:bool=True):
        new_u = self.u.y_rotation(angle, True)
        new_v = self.v.y_rotation(angle, True)
        new_w = self.w.y_rotation(angle, True)

        if copy:
            return Basis3D(new_u, new_v, new_w, self.name)
        else:
            self.u = new_u
            self.v = new_v
            self.w = new_w

    def z_rotation(self, angle:float, copy:bool=True):
        new_u = self.u.z_rotation(angle, True)
        new_v = self.v.z_rotation(angle, True)
        new_w = self.w.z_rotation(angle, True)

        if copy:
            return Basis3D(new_u, new_v, new_w, self.name)
        else:
            self.u = new_u
            self.v = new_v
            self.w = new_w

    def Eulerrotation(self, angles:Tuple[float, float, float], copy:bool=True):
        psi, theta, phi = angles
        center = O3D

        vect_u = self.u.copy()
        vect_v = self.v.copy()
        vect_w = self.w.copy()

        # rotation around w
        vect_u.rotation(center, vect_w, psi, False)
        vect_v.rotation(center, vect_w, psi, False)

        # rotation around v
        vect_v.rotation(center, vect_u, theta, False)
        vect_w.rotation(center, vect_u, theta, False)

        # rotation around w
        vect_u.rotation(center, vect_w, phi, False)
        vect_v.rotation(center, vect_w, phi, False)

        if copy:
            return Basis3D(vect_u, vect_v, vect_w)
        self.u = vect_u
        self.v = vect_v
        self.w = vect_w

    def transfer_matrix(self):
        return Matrix33(self.u.x, self.v.x, self.w.x,
                        self.u.y, self.v.y, self.w.y,
                        self.u.z, self.v.z, self.w.z)

    def inverse_transfer_matrix(self):
        return self.transfer_matrix().inverse()

    def new_coordinates(self, vector):
        matrix = self.inverse_transfer_matrix()
        return matrix.vector_multiplication(vector)

    def old_coordinates(self, point):
        matrix = self.transfer_matrix()
        return matrix.vector_multiplication(point)

    def copy(self):
        return Basis3D(self.u, self.v, self.w)

    def normalize(self):
        """
        normalize the basis modifying its coordinates in place
        """
        self.u.normalize()
        self.v.normalize()
        self.w.normalize()

class Frame2D(Basis2D):
    """
    Defines a 2D basis
    :param origin:Point2D: origin of the basis
    :param u:Vector2D: first vector of the basis
    :param v:Vector2D: second vector of the basis
    """
    def __init__(self, origin:Point2D, u:Vector2D, v:Vector2D, name:str=''):
        self.origin = origin
        Basis2D.__init__(self, u, v, name=name)

    def __repr__(self):
        return '{}: O={} U={}, V={}'.format(self.__class__.__name__, self.origin, self.u, self.v)

    def __neg__(self):
        Pinv = self.inverse_transfer_matrix()
        new_origin = Point2D(npy.dot(Pinv, self.origin))
        return Frame2D(new_origin,
                       Vector2D(Pinv[:, 0]),
                       Vector2D(Pinv[:, 1]))


    def __add__(self, other_frame):
        P1 = self.transfer_matrix()
        new_origin = Point2D(npy.dot(P1, other_frame.origin) + self.origin)
        M = npy.dot(P1, other_frame.transfer_matrix())
        return Frame2D(new_origin,
                       Vector2D(M[:, 0]),
                       Vector2D(M[:, 1]))


    def __sub__(self, other_frame):
        P1inv = other_frame.inverse_transfer_matrix()
        P2 = self.transfer_matrix()
        new_origin = Point2D(npy.dot(P1inv, (self.origin - other_frame.origin)))
        M = npy.dot(P1inv, P2)
        return Frame2D(new_origin,
                       Vector2D(M[:, 0]),
                       Vector2D(M[:, 1]))

    def basis(self):
        return Basis2D(self.u, self.v)

    def new_coordinates(self, vector):
        return Basis2D.new_coordinates(self, vector - self.origin)

    def old_coordinates(self, vector):
        return Basis2D.old_coordinates(self, vector) + self.origin

    def translation(self, vector, copy=True):
        new_origin = self.origin.translation(vector, True)
        if copy:
            return Frame2D(new_origin, self.u, self.v)
        self.origin = new_origin


    def rotation(self, angle, copy=True):
        new_base = Basis2D.rotation(self, angle, True)
        if copy:
            new_frame = Frame2D(self.origin, new_base.u, new_base.v)
            return new_frame
        self.u = new_base.u
        self.v = new_base.v

    def Draw(self, ax=None, style='ok'):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(*self.origin, style)
        self.u.plot(origin=self.origin, ax=ax, color='r')
        self.v.plot(origin=self.origin, ax=ax, color='g')
        ax.axis('equal')

    def copy(self):
        return Frame2D(self.origin, self.u, self.v)


OXY = Frame2D(O2D, X2D, Y2D)


class Frame3D(Basis3D):
    """
    Defines a 3D frame
    :param origin:Point3D: origin of the basis
    :param u:Vector3D: first vector of the basis
    :param v:Vector3D: second vector of the basis
    :param w:Vector3D: third vector of the basis
    """
    def __init__(self, origin:Point3D, u:Vector3D, v:Vector3D, w:Vector3D, name:str=''):
        self.origin = origin
        Basis3D.__init__(self, u, v, w)
        self.name = name

    def __repr__(self):
        return '{}: O={} U={}, V={}, W={}'.format(self.__class__.__name__,
                                                  self.origin,
                                                  self.u, self.v, self.w)


    def __neg__(self):
        M = self.inverse_transfer_matrix()
        new_origin = M.vector_multiplication(self.origin)
        return Frame3D(new_origin,
                       Vector3D(M.M11, M.M21, M.M31),
                       Vector3D(M.M12, M.M22, M.M32),
                       Vector3D(M.M13, M.M23, M.M33))


    def __add__(self, other_frame):
        P1 = self.transfer_matrix()
        new_origin = P1.vector_multiplication(other_frame.origin) + self.origin


        M = P1 * other_frame.transfer_matrix()
        return Frame3D(new_origin,
                       Vector3D(M.M11, M.M21, M.M31),
                       Vector3D(M.M12, M.M22, M.M32),
                       Vector3D(M.M13, M.M23, M.M33))


    def __sub__(self, other_frame):
        P1inv = other_frame.inverse_transfer_matrix()
        P2 = self.transfer_matrix()
        new_origin = P1inv.vector_multiplication(self.origin - other_frame.origin)
        M = P1inv * P2
        return Frame3D(new_origin,
                       Vector3D(M.M11, M.M21, M.M31),
                       Vector3D(M.M12, M.M22, M.M32),
                       Vector3D(M.M13, M.M23, M.M33))

    def __round__(self, ndigits=6):
        return self.__class__(round(self.origin, ndigits),
                              round(self.u, ndigits),
                              round(self.v, ndigits),
                              round(self.w, ndigits))

    def __hash__(self):
        return hash(self.u) + hash(self.v) + hash(self.w) + hash(self.origin)

    def basis(self):
        return Basis3D(self.u, self.v, self.w)

    def new_coordinates(self, vector):
        """ You have to give coordinates in the global landmark """
        return Basis3D.new_coordinates(self, vector - self.origin)

    def old_coordinates(self, vector):
        """ You have to give coordinates in the local landmark """
        return Basis3D.old_coordinates(self, vector) + self.origin

    def rotation(self, axis, angle, copy=True):
        new_base = Basis3D.rotation(self, axis, angle, copy=True)
        if copy:
            new_frame = Frame3D(self.origin.copy(), new_base.u, new_base.v, new_base.w, self.name)
            return new_frame
        self.u = new_base.u
        self.v = new_base.v
        self.w = new_base.w

    def translation(self, offset, copy=True):
        if copy:
            return Frame3D(self.origin.translation(offset, copy=True), self.u, self.v, self.w, self.name)
        self.origin.translation(offset, copy=False)

    def copy(self):
        return Frame3D(self.origin.copy(), self.u.copy(), self.v.copy(), self.w.copy())

    def to_step(self, current_id):

        content, origin_id = self.origin.to_point().to_step(current_id)
        current_id = origin_id + 1
        u_content, u_id = Vector3D.to_step(self.u, current_id)
        current_id = u_id + 1
        v_content, v_id = Vector3D.to_step(self.v, current_id)
        current_id = v_id + 1
        content += u_content + v_content
        content += "#{} = AXIS2_PLACEMENT_3D('{}',#{},#{},#{});\n"\
                        .format(current_id, self.name, origin_id, u_id, v_id)
        return content, current_id


    def plot2d(self, x=X3D, y=Y3D, ax=None, color='k'):
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        origin2d = self.origin.to_2d(O3D, x, y)

        for iv, vector in enumerate(self.vectors):
            vector2D = vector.to_2d(O3D, x, y)
            if vector2D.norm() > 1e-8:
                vector2D.plot(origin=origin2d, ax=ax, color=color, label=str(iv+1))

        return fig, ax


    def plot(self, ax=None, color='b', alpha=1., plot_points=True):
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

        x1 = [p.x for p in (self.origin, self.origin + self.u)]
        y1 = [p.y for p in (self.origin, self.origin + self.u)]
        z1 = [p.z for p in (self.origin, self.origin + self.u)]
        ax.plot(x1, y1, z1, 'r')

        x2 = [p.x for p in (self.origin, self.origin + self.v)]
        y2 = [p.y for p in (self.origin, self.origin + self.v)]
        z2 = [p.z for p in (self.origin, self.origin + self.v)]
        ax.plot(x2, y2, z2, 'g')

        x3 = [p.x for p in (self.origin, self.origin + self.w)]
        y3 = [p.y for p in (self.origin, self.origin + self.w)]
        z3 = [p.z for p in (self.origin, self.origin + self.w)]
        ax.plot(x3, y3, z3, 'b')
        return ax

    @classmethod
    def from_step(cls, arguments, object_dict):
        origin = object_dict[arguments[1]]
        if arguments[2] == '$':
            u = None
        else:
            u = object_dict[arguments[2]]
        if arguments[3] == '$':
            v = None
        else:
            v = object_dict[arguments[3]]
        if u is None or v is None:
            w = None
        else:
            w = u.cross(v)

        return cls(origin, u, v, w, arguments[0][1:-1])


    def babylonjs(self, size=0.1, parent=None):
        s = 'var origin = new BABYLON.Vector3({},{},{});\n'.format(*self.origin)
        s += 'var o_u = new BABYLON.Vector3({}, {}, {});\n'.format(*(size*self.u+self.origin))
        s += 'var o_v = new BABYLON.Vector3({}, {}, {});\n'.format(*(size*self.v+self.origin))
        s += 'var o_w = new BABYLON.Vector3({}, {}, {});\n'.format(*(size*self.w+self.origin))
        s += 'var line1 = BABYLON.MeshBuilder.CreateTube("frame_U", {{path: [origin, o_u], radius: {}}}, scene);'.format(0.03*size)
        s += 'line1.material = red_material;\n'
        s += 'var line2 = BABYLON.MeshBuilder.CreateTube("frame_V", {{path: [origin, o_v], radius: {}}}, scene);'.format(0.03*size)
        s += 'line2.material = green_material;\n'
        s += 'var line3 = BABYLON.MeshBuilder.CreateTube("frame_W", {{path: [origin, o_w], radius: {}}}, scene);'.format(0.03*size)
        s += 'line3.material = blue_material;\n'
        if parent is not None:
            s += 'line1.parent = {};\n'.format(parent)
            s += 'line2.parent = {};\n'.format(parent)
            s += 'line3.parent = {};\n'.format(parent)

        return s




# -*- coding: utf-8 -*-

import math
import pkg_resources
__version__ = pkg_resources.require("volmdlr")[0].version

from volmdlr.core_compiled import (Vector2D, Vector3D, Point2D, Point3D,
                            O2D, X2D, Y2D, OXY,
                            Basis2D, Basis3D, Frame2D, Frame3D,
                            O3D, X3D, Y3D, Z3D, Matrix22, Matrix33
                            )

TWO_PI = 2*math.pi

XYZ = Basis3D(X3D, Y3D, Z3D)
YZX = Basis3D(Y3D, Z3D, X3D)
ZXY = Basis3D(Z3D, X3D, Y3D)

OXYZ = Frame3D(O3D, X3D, Y3D, Z3D)
OYZX = Frame3D(O3D, Y3D, Z3D, X3D)
OZXY = Frame3D(O3D, Z3D, X3D, Y3D)



import os
import pkg_resources
from string import Template


babylon_unpacker_cdn_header = '''
<!doctype html>
<html>
<head>
   <meta charset="utf-8">
   <title>Babylon from volmdlr</title>
   <style>
      html, body {
         overflow: hidden;
         width: 100%;
         height: 100%;
         margin: 0;
         padding: 0;
      }
      #renderCanvas {
         width: 100%;
         height: 100%;
         touch-action: none;
      }
   </style>
      <script src="https://preview.babylonjs.com/babylon.js"></script>
      <script src="https://preview.babylonjs.com/loaders/babylonjs.loaders.min.js"></script>
      <script src="https://code.jquery.com/pep/0.4.3/pep.js"></script>
      <script src='https://unpkg.com/earcut@2.1.1/dist/earcut.min.js'></script>
      <script src='https://preview.babylonjs.com/gui/babylon.gui.min.js'></script>
</head>
'''

babylon_unpacker_embedded_header = '''
<!doctype html>
<html>
<head>
   <meta charset="utf-8">
   <title>Babylon from volmdlr</title>
   <style>
      html, body {
         overflow: hidden;
         width: 100%;
         height: 100%;
         margin: 0;
         padding: 0;
      }
      #renderCanvas {
         width: 100%;
         height: 100%;
         touch-action: none;
      }
   </style>
   <script>
   '''

for filename in ['babylon.js', 'babylonjs.loaders.min.js', 'earcut.min.js', 'pep.js']:
    with pkg_resources.resource_stream(
            pkg_resources.Requirement('volmdlr'),
            os.path.join('volmdlr/assets/js/', filename)) as fjs:
        babylon_unpacker_embedded_header += fjs.read().decode('utf-8')

babylon_unpacker_embedded_header += '''
      </script>
</head>
'''


babylon_unpacker_body_template = Template(
'''
<body>
   <canvas id="renderCanvas"></canvas>
   <script type="text/javascript">
      // Get the canvas element from our HTML below
      var canvas = document.querySelector("#renderCanvas");
      // Load the BABYLON 3D engine
      var engine = new BABYLON.Engine(canvas, true);

      var babylon_data = $babylon_data;
      var max_length = babylon_data['max_length'];

      // -------------------------------------------------------------
      // Here begins a function that we will 'call' just after it's built
      var createScene = function () {
        // This creates a basic Babylon Scene object (non-mesh)
        var scene = new BABYLON.Scene(engine);
        scene.useRightHandedSystem = true;
        scene.clearColor = new BABYLON.Color4(.9, .9, .9, .9);
      	var cam = new BABYLON.ArcRotateCamera("ArcRotateCamera",
                                              0, 0, 2*babylon_data['max_length'],
                                              new BABYLON.Vector3(babylon_data['center'][0],
                                                                  babylon_data['center'][1],
                                                                  babylon_data['center'][2]), scene);
      	cam.wheelPrecision=50./babylon_data['max_length']
      	cam.pinchPrecision=50./babylon_data['max_length']
      	cam.panningSensibility=800./babylon_data['max_length'];
      	cam.minZ=0.01*babylon_data['max_length'];
      	cam.attachControl(canvas);
      	cam.inertia = 0;
      	cam.panningInertia = 0;
      	// cam.mode = BABYLON.Camera.ORTHOGRAPHIC_CAMERA;
      	cam.upVector = new BABYLON.Vector3(0, 0, 1);
      	cam.lowerBetaLimit = null;
        cam.upperBetaLimit = null;
        cam.checkCollisions = false;


      	var light1 = new BABYLON.HemisphericLight("light1", new BABYLON.Vector3(-1, -1, -1), scene);
      	light1.intensity=0.5;
      	light1.specular = new BABYLON.Color3(0, 0, 0);

      	// var light2 = new BABYLON.SpotLight("Spot0", new BABYLON.Vector3(0, 30, -10), new BABYLON.Vector3(0, -1, 0), 0.8, 2, scene);
      	// light2.diffuse = new BABYLON.Color3(1, 1, 1);
      	// light2.specular = new BABYLON.Color3(1, 1, 1);
        var light2 = new BABYLON.PointLight("light2", new BABYLON.Vector3(0, 0, 0), scene);
        light2.specular = new BABYLON.Color3(0, 0, 0);
        light2.intensity = 0.3;
        light2.parent = cam;

        var light3 = new BABYLON.HemisphericLight("light3", new BABYLON.Vector3(1, 1, 1), scene);
        light3.specular = new BABYLON.Color3(0, 0, 0);
        light3.intensity = 0.50;


        var showAxis = function (size) {
            var makeTextPlane = function (text, color, size) {
                var dynamicTexture = new BABYLON.DynamicTexture("DynamicTexture", 50, scene, true);
                dynamicTexture.hasAlpha = true;
                dynamicTexture.drawText(text, 5, 40, "bold 36px Arial", color, "transparent", true);
                var plane = new BABYLON.Mesh.CreatePlane("TextPlane", size, scene, true);
                plane.material = new BABYLON.StandardMaterial("TextPlaneMaterial", scene);
                plane.material.backFaceCulling = false;
                plane.material.specularColor = new BABYLON.Color3(0, 0, 0);
                plane.material.diffuseTexture = dynamicTexture;
                return plane;
            };
            var axisX = BABYLON.Mesh.CreateLines("axisX", [
                new BABYLON.Vector3.Zero(), new BABYLON.Vector3(size, 0, 0), new BABYLON.Vector3(size * 0.95, 0.05 * size, 0),
                new BABYLON.Vector3(size, 0, 0), new BABYLON.Vector3(size * 0.95, -0.05 * size, 0)
            ], scene);
            axisX.color = new BABYLON.Color3(1, 0, 0);
            var xChar = makeTextPlane("X", "red", size / 10);
            xChar.position = new BABYLON.Vector3(0.9 * size, -0.05 * size, 0);
            var axisY = BABYLON.Mesh.CreateLines("axisY", [
                new BABYLON.Vector3.Zero(), new BABYLON.Vector3(0, size, 0), new BABYLON.Vector3(-0.05 * size, size * 0.95, 0),
                new BABYLON.Vector3(0, size, 0), new BABYLON.Vector3(0.05 * size, size * 0.95, 0)
            ], scene);
            axisY.color = new BABYLON.Color3(0, 1, 0);
            var yChar = makeTextPlane("Y", "green", size / 10);
            yChar.position = new BABYLON.Vector3(0, 0.9 * size, -0.05 * size);
            var axisZ = BABYLON.Mesh.CreateLines("axisZ", [
                new BABYLON.Vector3.Zero(), new BABYLON.Vector3(0, 0, size), new BABYLON.Vector3(0, -0.05 * size, size * 0.95),
                new BABYLON.Vector3(0, 0, size), new BABYLON.Vector3(0, 0.05 * size, size * 0.95)
            ], scene);
            axisZ.color = new BABYLON.Color3(0, 0, 1);
            var zChar = makeTextPlane("Z", "blue", size / 10);
            zChar.position = new BABYLON.Vector3(0, 0.05 * size, 0.9 * size);
        };

        showAxis(1);


        var meshes = []
        for (let mesh_data of babylon_data['meshes']){
          var mesh = new BABYLON.Mesh(mesh_data['name'], scene);
          meshes.push(mesh);

          var normals = [];
          var vertexData = new BABYLON.VertexData();
          BABYLON.VertexData.ComputeNormals(mesh_data['positions'], mesh_data['indices'], normals);

          vertexData.positions = mesh_data['positions'];
          vertexData.indices = mesh_data['indices'];
          vertexData.normals = normals;
          vertexData.applyToMesh(mesh);
          mesh.enableEdgesRendering(0.9);
          mesh.edgesWidth = max_length*0.3;
          mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 0.6);
          var mat = new BABYLON.StandardMaterial("material", scene);
          // mat.diffuseColor = BABYLON.Color3.Green();
          // mat.specularColor = new BABYLON.Color3(0.5, 0.6, 0.87);
          // mat.emissiveColor = new BABYLON.Color3(1, 1, 1);
          // mat.ambientColor = new BABYLON.Color3(0.23, 0.98, 0.53);
          mat.backFaceCulling = false;
          mesh.material = mat;
          mat.diffuseColor = new BABYLON.Color3(mesh_data['color'][0],
                                                mesh_data['color'][1],
                                                mesh_data['color'][2]);
          mat.alpha = mesh_data['alpha'];

        }


        if (babylon_data['steps']){


          var n_primitives = babylon_data['meshes'].length;
          var n_steps = babylon_data['steps'].length;

          var advancedTexture = BABYLON.GUI.AdvancedDynamicTexture.CreateFullscreenUI("UI");
          var animation_stopped = false;

          var buttonHeightInPixels = 50;
          var buttonWidthInPixels = 150;

          // Set container and its size and position
          let heightInPixels = buttonHeightInPixels*7;
          let widthInPixels = buttonWidthInPixels;

          let topInPixels = -canvas.height/2 + heightInPixels/2;
          let leftInPixels = -canvas.width/2 + widthInPixels/2;

          var buttonsContainer = new BABYLON.GUI.StackPanel("buttons_panel");
          buttonsContainer.background = "#263238";
          buttonsContainer.color = "white";
          buttonsContainer.height = ""+heightInPixels+"px";
          buttonsContainer.width = ""+widthInPixels+"px";
          buttonsContainer.top = ""+topInPixels+"px";
          buttonsContainer.left = ""+leftInPixels+"px";

          var step_number_label = new BABYLON.GUI.TextBlock();
          step_number_label.text = "Step n°1/"+n_steps;
          step_number_label.width = ""+buttonWidthInPixels+"px";
          step_number_label.height = ""+buttonHeightInPixels+"px";
          buttonsContainer.addControl(step_number_label);


          var start_stop_button = BABYLON.GUI.Button.CreateSimpleButton("animation", "Stop/Resume animation");
          start_stop_button.width = ""+buttonWidthInPixels+"px";
          start_stop_button.height = ""+buttonHeightInPixels+"px";
          start_stop_button.onPointerUpObservable.add(function(){animation_stopped = !animation_stopped});
          buttonsContainer.addControl(start_stop_button);

          var first_step_button = BABYLON.GUI.Button.CreateSimpleButton("animation", "First step");
          first_step_button.width = ""+buttonWidthInPixels+"px";
          first_step_button.height = ""+buttonHeightInPixels+"px";
          first_step_button.onPointerUpObservable.add(function(){animation_stopped=true; iframe=0; showStep(Math.floor(0))});
          buttonsContainer.addControl(first_step_button);

          var previous_step_button = BABYLON.GUI.Button.CreateSimpleButton("animation", "Previous step");
          previous_step_button.width = ""+buttonWidthInPixels+"px";
          previous_step_button.height = ""+buttonHeightInPixels+"px";
          previous_step_button.onPointerUpObservable.add(function(){animation_stopped=true; iframe-=10; showStep(Math.floor(iframe/10))});
          buttonsContainer.addControl(previous_step_button);

          var next_step_button = BABYLON.GUI.Button.CreateSimpleButton("animation", "Next step");
          next_step_button.width = ""+buttonWidthInPixels+"px";
          next_step_button.height = ""+buttonHeightInPixels+"px";
          next_step_button.onPointerUpObservable.add(function(){animation_stopped=true; iframe+=10; showStep(Math.floor(iframe/10))});
          buttonsContainer.addControl(next_step_button);

          var last_step_button = BABYLON.GUI.Button.CreateSimpleButton("animation", "Last step");
          last_step_button.width = ""+buttonWidthInPixels+"px";
          last_step_button.height = ""+buttonHeightInPixels+"px";
          last_step_button.onPointerUpObservable.add(function(){animation_stopped=true; iframe=10*(n_steps-1); showStep(Math.floor(iframe/10))});
          buttonsContainer.addControl(last_step_button);

          var step_label = new BABYLON.GUI.TextBlock();
          step_label.text = "Label";
          step_label.width = ""+buttonWidthInPixels+"px";
          step_label.height = ""+buttonHeightInPixels+"px";
          buttonsContainer.addControl(step_label);

          advancedTexture.addControl(buttonsContainer);


          var showStep = function (istep){
            // Parts position update
            step_number_label.text = "Step n°"+(istep+1)+"/"+n_steps;
            for(let iprimitive=0; iprimitive<n_primitives; iprimitive++){
                mesh = meshes[iprimitive];
                mesh.position = new BABYLON.Vector3(babylon_data['steps'][istep][iprimitive]['position'][0],
                                                    babylon_data['steps'][istep][iprimitive]['position'][1],
                                                    babylon_data['steps'][istep][iprimitive]['position'][2]);


                mesh.rotation = BABYLON.Vector3.RotationFromAxis(
                  new BABYLON.Vector3(babylon_data['steps'][istep][iprimitive]['orientations'][0][0],
                                      babylon_data['steps'][istep][iprimitive]['orientations'][0][1],
                                      babylon_data['steps'][istep][iprimitive]['orientations'][0][2]),
                  new BABYLON.Vector3(babylon_data['steps'][istep][iprimitive]['orientations'][1][0],
                                      babylon_data['steps'][istep][iprimitive]['orientations'][1][1],
                                      babylon_data['steps'][istep][iprimitive]['orientations'][1][2]),
                  new BABYLON.Vector3(babylon_data['steps'][istep][iprimitive]['orientations'][2][0],
                                      babylon_data['steps'][istep][iprimitive]['orientations'][2][1],
                                      babylon_data['steps'][istep][iprimitive]['orientations'][2][2]));
            }

            if ('label' in babylon_data['steps'][istep]){
              step_label.text = babylon_data['steps'][istep]['label'];
            }
            else{
              step_label.text = "";
            }

          }

        var iframe = 0;

        scene.registerBeforeRender(function () {
                  if (!animation_stopped){
                    if (iframe % 10 == 0){
                      var istep = iframe / 10;

                      showStep(istep);


                    }
                    iframe++;
                    iframe = iframe % (n_steps*10);
                  }

          });
          }


    	return scene;	  };

      var scene = createScene();

      // Register a render loop to repeatedly render the scene
      engine.runRenderLoop(function () {
         scene.render();
      });
      // Watch for browser/canvas resize events
      window.addEventListener("resize", function () {
         engine.resize();
      });

        //scene.debugLayer.show();

   </script>
</body>

</html>

'''
)